import itertools
from enum import IntEnum
from mstk import *
from mstk.simsys import System
from mstk.forcefield import *
from mstk.topology import *
from mstk.chem.constant import *

try:
    import openmm.openmm as mm
except ImportError:
    OPENMM_FOUND = False
else:
    OPENMM_FOUND = True


class ForceGroup(IntEnum):
    BOND = 1
    ANGLE = 2
    DIHEDRAL = 3
    IMPROPER = 4
    VDW = 5
    COULOMB = 6
    DRUDE = 7


class OpenMMExporter:
    '''
    OpenMMExporter export a :class:`System` to a OpenMM system.

    All of the force field terms supported by `mstk` can be exported to OpenMM.
    '''

    def __init__(self):
        pass

    @staticmethod
    def export(system, disable_inter_mol=False, **kwargs):
        '''
        Generate OpenMM system from a system

        Parameters
        ----------
        system : System
        disable_inter_mol : bool

        Returns
        -------
        omm_system : openmm.openmm.System
        '''
        if not OPENMM_FOUND:
            raise ImportError('Can not import OpenMM')

        supported_terms = {LJ126Term, MieTerm,
                           HarmonicBondTerm, MorseBondTerm,
                           HarmonicAngleTerm, SDKAngleTerm, LinearAngleTerm,
                           OplsDihedralTerm, PeriodicDihedralTerm,
                           OplsImproperTerm, HarmonicImproperTerm,
                           DrudeTerm}
        unsupported = system.ff_classes - supported_terms
        if unsupported != set():
            raise Exception('Unsupported FF terms: %s'
                            % (', '.join(map(lambda x: x.__name__, unsupported))))

        if system.vsite_types - {TIP4PSite} != set():
            raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')

        omm_system = mm.System()
        if system.use_pbc:
            omm_system.setDefaultPeriodicBoxVectors(*system.topology.cell.vectors)
        for atom in system.topology.atoms:
            omm_system.addParticle(atom.mass)

        OpenMMExporter._setup_bonded(system, omm_system)
        if not disable_inter_mol:
            OpenMMExporter._setup_nonbonded(system, omm_system)
        else:
            OpenMMExporter._setup_nonbonded_as_bonded(system, omm_system)

        ### Remove COM motion ###################################################################
        logger.debug('Setting up COM motion remover...')
        omm_system.addForce(mm.CMMotionRemover(10))

        return omm_system

    @staticmethod
    def _setup_bonded(system, omm_system):
        top = system.topology
        ff = system.ff

        ### Set up bonds #######################################################################
        for bond_class in system.bond_classes:
            if bond_class == HarmonicBondTerm:
                logger.debug('Setting up harmonic bonds...')
                bforce = mm.HarmonicBondForce()
                for bond in top.bonds:
                    if bond.is_drude:
                        # DrudeForce will handle the bond between Drude pair
                        continue
                    if bond in system.constrain_bonds:
                        continue
                    bterm = system.bond_terms[bond]
                    if type(bterm) != HarmonicBondTerm:
                        continue
                    bforce.addBond(bond.atom1.id, bond.atom2.id, bterm.length, bterm.k * 2)
            elif bond_class == MorseBondTerm:
                logger.debug('Setting up Morse bonds...')
                bforce = mm.CustomBondForce('depth*(1-exp(-alpha*(r-b0)))^2;'
                                            'alpha=(k/depth)^0.5')
                bforce.addPerBondParameter('b0')
                bforce.addPerBondParameter('k')
                bforce.addPerBondParameter('depth')
                for bond in top.bonds:
                    if bond.is_drude:
                        # DrudeForce will handle the bond between Drude pair
                        continue
                    if bond in system.constrain_bonds:
                        continue
                    bterm = system.bond_terms[bond]
                    if type(bterm) != MorseBondTerm:
                        continue
                    bforce.addBond(bond.atom1.id, bond.atom2.id, [bterm.length, bterm.k, bterm.depth])
            else:
                raise Exception('Bond terms other that HarmonicBondTerm and MorseBondTerm '
                                'haven\'t been implemented')
            bforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
            bforce.setForceGroup(ForceGroup.BOND)
            bforce.setName('Bond')
            omm_system.addForce(bforce)

        ### Set up angles #######################################################################
        for angle_class in system.angle_classes:
            if angle_class == HarmonicAngleTerm:
                logger.debug('Setting up harmonic angles...')
                aforce = mm.HarmonicAngleForce()
                for angle in top.angles:
                    if angle in system.constrain_angles:
                        continue
                    aterm = system.angle_terms[angle]
                    if type(aterm) == HarmonicAngleTerm:
                        aforce.addAngle(angle.atom1.id, angle.atom2.id, angle.atom3.id,
                                        aterm.theta, aterm.k * 2)
            elif angle_class == SDKAngleTerm:
                logger.debug('Setting up SDK angles...')
                aforce = mm.CustomCompoundBondForce(
                    3, 'k*(theta-theta0)^2+step(rmin-r)*LJ;'
                       'LJ=C*epsilon*((sigma/r)^n-(sigma/r)^m)+epsilon;'
                       'theta=angle(p1,p2,p3);'
                       'r=distance(p1,p3);'
                       'C=n/(n-m)*(n/m)^(m/(n-m));'
                       'rmin=(n/m)^(1/(n-m))*sigma')
                aforce.addPerBondParameter('theta0')
                aforce.addPerBondParameter('k')
                aforce.addPerBondParameter('epsilon')
                aforce.addPerBondParameter('sigma')
                aforce.addPerBondParameter('n')
                aforce.addPerBondParameter('m')
                for angle in top.angles:
                    if angle in system.constrain_angles:
                        continue
                    aterm = system.angle_terms[angle]
                    if type(aterm) != SDKAngleTerm:
                        continue
                    vdw = ff.get_vdw_term(ff.atom_types[angle.atom1.type], ff.atom_types[angle.atom3.type])
                    if type(vdw) != MieTerm:
                        raise Exception(f'Corresponding MieTerm for {aterm} not found in FF')
                    aforce.addBond([angle.atom1.id, angle.atom2.id, angle.atom3.id],
                                   [aterm.theta, aterm.k, vdw.epsilon, vdw.sigma, vdw.repulsion, vdw.attraction])
            elif angle_class == LinearAngleTerm:
                logger.debug('Setting up linear angles...')
                aforce = mm.CustomCompoundBondForce(
                    3, 'k_linear*r^2;'
                       'r=pointdistance(xref,yref,zref,x2,y2,z2);'
                       'xref=x1*a+x3*(1-a);'
                       'yref=y1*a+y3*(1-a);'
                       'zref=z1*a+z3*(1-a)')
                aforce.addPerBondParameter('a')
                aforce.addPerBondParameter('k_linear')
                for angle in top.angles:
                    if angle in system.constrain_angles:
                        continue
                    aterm = system.angle_terms[angle]
                    if type(aterm) != LinearAngleTerm:
                        continue
                    bond12, bond23 = angle.bonds
                    bterm12 = system.bond_terms[bond12]
                    bterm23 = system.bond_terms[bond23]
                    a, k_linear = aterm.calc_a_k_linear(bterm12.length, bterm23.length)
                    aforce.addBond([angle.atom1.id, angle.atom2.id, angle.atom3.id], [a, k_linear])
            else:
                raise Exception('Angle terms other that HarmonicAngleTerm, SDKAngleTerm and LinearAngleTerm '
                                'haven\'t been implemented')
            aforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
            aforce.setForceGroup(ForceGroup.ANGLE)
            aforce.setName('Angle')
            omm_system.addForce(aforce)

        ### Set up constraints #################################################################
        logger.debug(f'Setting up {len(system.constrain_bonds)} bond constraints...')
        for bond in top.bonds:
            if bond in system.constrain_bonds:
                omm_system.addConstraint(bond.atom1.id, bond.atom2.id,
                                         system.constrain_bonds[bond])
        logger.debug(f'Setting up {len(system.constrain_angles)} angle constraints...')
        for angle in top.angles:
            if angle in system.constrain_angles:
                omm_system.addConstraint(angle.atom1.id, angle.atom3.id,
                                         system.constrain_angles[angle])

        ### Set up dihedrals ###################################################################
        for dihedral_class in system.dihedral_classes:
            if dihedral_class in (OplsDihedralTerm, PeriodicDihedralTerm):
                logger.debug('Setting up periodic dihedrals...')
                dforce = mm.PeriodicTorsionForce()
                for dihedral in top.dihedrals:
                    dterm = system.dihedral_terms[dihedral]
                    ia1, ia2, ia3, ia4 = dihedral.atom1.id, dihedral.atom2.id, dihedral.atom3.id, dihedral.atom4.id
                    if type(dterm) is OplsDihedralTerm:
                        dterm = dterm.to_periodic_term()
                    if type(dterm) == PeriodicDihedralTerm:
                        for phase in dterm.phases:
                            if phase.k != 0:
                                dforce.addTorsion(ia1, ia2, ia3, ia4, phase.n, phase.phi, phase.k)
            else:
                raise Exception('Dihedral terms other that OplsDihedralTerm and PeriodicDihedralTerm '
                                'haven\'t been implemented')
            dforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
            dforce.setForceGroup(ForceGroup.DIHEDRAL)
            dforce.setName('Dihedral')
            omm_system.addForce(dforce)

        ### Set up impropers ####################################################################
        for improper_class in system.improper_classes:
            if improper_class == OplsImproperTerm:
                logger.debug('Setting up periodic impropers...')
                iforce = mm.CustomTorsionForce('k*(1-cos(2*theta))')
                iforce.addPerTorsionParameter('k')
                for improper in top.impropers:
                    iterm = system.improper_terms[improper]
                    if type(iterm) == OplsImproperTerm:
                        # in OPLS convention, the third atom is the central atom
                        iforce.addTorsion(improper.atom2.id, improper.atom3.id,
                                          improper.atom1.id, improper.atom4.id, [iterm.k])
            elif improper_class == HarmonicImproperTerm:
                logger.debug('Setting up harmonic impropers...')
                iforce = mm.CustomTorsionForce(f'k*min(dtheta,2*{PI}-dtheta)^2;'
                                               f'dtheta=abs(theta-phi0)')
                iforce.addPerTorsionParameter('phi0')
                iforce.addPerTorsionParameter('k')
                for improper in top.impropers:
                    iterm = system.improper_terms[improper]
                    if type(iterm) == HarmonicImproperTerm:
                        iforce.addTorsion(improper.atom1.id, improper.atom2.id,
                                          improper.atom3.id, improper.atom4.id,
                                          [iterm.phi, iterm.k])
            else:
                raise Exception('Improper terms other that PeriodicImproperTerm and '
                                'HarmonicImproperTerm haven\'t been implemented')
            iforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
            iforce.setForceGroup(ForceGroup.IMPROPER)
            iforce.setName('Improper')
            omm_system.addForce(iforce)

    @staticmethod
    def _setup_nonbonded(system, omm_system):
        top = system.topology
        ff = system.ff

        ### Set up non-bonded interactions #########################################################
        # NonbonedForce is not flexible enough. Use it only for Coulomb interactions (including 1-4 Coulomb exceptions)
        # CustomNonbondedForce handles vdW interactions
        # CustomBondForce handles 1-4 vdW exceptions
        cutoff = ff.vdw_cutoff
        logger.debug('Setting up Coulomb interactions...')
        nbforce = mm.NonbondedForce()
        if system.use_pbc:
            nbforce.setNonbondedMethod(mm.NonbondedForce.PME)
            nbforce.setEwaldErrorTolerance(5E-4)
            nbforce.setCutoffDistance(cutoff)
            # dispersion will be handled by CustomNonbondedForce
            nbforce.setUseDispersionCorrection(False)
            nbforce.setExceptionsUsePeriodicBoundaryConditions(True)
        else:
            nbforce.setNonbondedMethod(mm.NonbondedForce.NoCutoff)
        nbforce.setForceGroup(ForceGroup.COULOMB)
        nbforce.setName('Coulomb')
        for atom in top.atoms:
            nbforce.addParticle(atom.charge, 1.0, 0.0)
        if system.charged:
            omm_system.addForce(nbforce)

        ### Set up vdW interactions #########################################################
        atom_types = list(ff.atom_types.values())
        type_names = list(ff.atom_types.keys())
        n_type = len(atom_types)
        for vdw_class in system.vdw_classes:
            if vdw_class == LJ126Term:
                logger.debug('Setting up LJ-12-6 vdW interactions...')
                if system.use_pbc and ff.vdw_long_range == ForceField.VDW_LONGRANGE_SHIFT:
                    invRc6 = 1 / cutoff ** 6
                    cforce = mm.CustomNonbondedForce(
                        f'A(type1,type2)*(invR6*invR6-{invRc6 * invRc6})-'
                        f'B(type1,type2)*(invR6-{invRc6});'
                        f'invR6=1/r^6')
                else:
                    cforce = mm.CustomNonbondedForce(
                        'A(type1,type2)*invR6*invR6-B(type1,type2)*invR6;'
                        'invR6=1/r^6')
                cforce.addPerParticleParameter('type')
                A_list = [0.0] * n_type * n_type
                B_list = [0.0] * n_type * n_type
                for i, atype1 in enumerate(atom_types):
                    for j, atype2 in enumerate(atom_types):
                        vdw = ff.get_vdw_term(atype1, atype2)
                        if type(vdw) == LJ126Term:
                            A = 4 * vdw.epsilon * vdw.sigma ** 12
                            B = 4 * vdw.epsilon * vdw.sigma ** 6
                        else:
                            A = B = 0
                        A_list[i + n_type * j] = A
                        B_list[i + n_type * j] = B
                cforce.addTabulatedFunction('A', mm.Discrete2DFunction(n_type, n_type, A_list))
                cforce.addTabulatedFunction('B', mm.Discrete2DFunction(n_type, n_type, B_list))

                for atom in top.atoms:
                    id_type = type_names.index(atom.type)
                    cforce.addParticle([id_type])

            elif vdw_class == MieTerm:
                logger.debug('Setting up Mie vdW interactions...')
                if system.use_pbc and ff.vdw_long_range == ForceField.VDW_LONGRANGE_SHIFT:
                    cforce = mm.CustomNonbondedForce('A(type1,type2)/r^REP(type1,type2)-'
                                                     'B(type1,type2)/r^ATT(type1,type2)-'
                                                     'SHIFT(type1,type2)')
                else:
                    cforce = mm.CustomNonbondedForce('A(type1,type2)/r^REP(type1,type2)-'
                                                     'B(type1,type2)/r^ATT(type1,type2)')
                cforce.addPerParticleParameter('type')
                A_list = [0.0] * n_type * n_type
                B_list = [0.0] * n_type * n_type
                REP_list = [0.0] * n_type * n_type
                ATT_list = [0.0] * n_type * n_type
                SHIFT_list = [0.0] * n_type * n_type
                for i, atype1 in enumerate(atom_types):
                    for j, atype2 in enumerate(atom_types):
                        vdw = ff.get_vdw_term(atype1, atype2)
                        if type(vdw) == MieTerm:
                            A = vdw.factor_energy() * vdw.epsilon * vdw.sigma ** vdw.repulsion
                            B = vdw.factor_energy() * vdw.epsilon * vdw.sigma ** vdw.attraction
                            REP = vdw.repulsion
                            ATT = vdw.attraction
                            SHIFT = A / cutoff ** REP - B / cutoff ** ATT
                        else:
                            A = B = REP = ATT = SHIFT = 0
                        A_list[i + n_type * j] = A
                        B_list[i + n_type * j] = B
                        REP_list[i + n_type * j] = REP
                        ATT_list[i + n_type * j] = ATT
                        SHIFT_list[i + n_type * j] = SHIFT
                cforce.addTabulatedFunction('A', mm.Discrete2DFunction(n_type, n_type, A_list))
                cforce.addTabulatedFunction('B', mm.Discrete2DFunction(n_type, n_type, B_list))
                cforce.addTabulatedFunction('REP', mm.Discrete2DFunction(n_type, n_type, REP_list))
                cforce.addTabulatedFunction('ATT', mm.Discrete2DFunction(n_type, n_type, ATT_list))
                if system.use_pbc and ff.vdw_long_range == ForceField.VDW_LONGRANGE_SHIFT:
                    cforce.addTabulatedFunction('SHIFT', mm.Discrete2DFunction(n_type, n_type, SHIFT_list))

                for atom in top.atoms:
                    id_type = type_names.index(atom.type)
                    cforce.addParticle([id_type])

            else:
                raise Exception('vdW terms other than LJ126Term and MieTerm '
                                'haven\'t been implemented')
            if system.use_pbc:
                cforce.setNonbondedMethod(mm.CustomNonbondedForce.CutoffPeriodic)
                cforce.setCutoffDistance(cutoff)
                if ff.vdw_long_range == ForceField.VDW_LONGRANGE_CORRECT:
                    cforce.setUseLongRangeCorrection(True)
            else:
                cforce.setNonbondedMethod(mm.CustomNonbondedForce.NoCutoff)
            cforce.setForceGroup(ForceGroup.VDW)
            cforce.setName('vdW')
            omm_system.addForce(cforce)

        ### Set up 1-2, 1-3 and 1-4 exceptions ##################################################
        logger.debug('Setting up 1-2, 1-3 and 1-4 exceptions...')
        custom_nb_forces = [f for f in omm_system.getForces() if type(f) == mm.CustomNonbondedForce]
        pair12, pair13, pair14 = top.get_12_13_14_pairs()
        for atom1, atom2 in pair12 + pair13:
            nbforce.addException(atom1.id, atom2.id, 0.0, 1.0, 0.0)
            for f in custom_nb_forces:
                f.addExclusion(atom1.id, atom2.id)
        # As long as 1-4 LJ OR Coulomb need to be scaled, then this pair should be excluded from ALL non-bonded forces.
        # This is required by OpenMM's internal implementation.
        # Even though NonbondedForce can handle 1-4 vdW, we use it only for 1-4 Coulomb.
        # And use CustomBondForce to handle 1-4 vdW, which makes it more clear for energy decomposition.
        if ff.scale_14_vdw != 1 or ff.scale_14_coulomb != 1:
            pair14_forces = {}  # {VdwTerm: mm.NbForce}
            for atom1, atom2 in pair14:
                charge_prod = atom1.charge * atom2.charge * ff.scale_14_coulomb
                nbforce.addException(atom1.id, atom2.id, charge_prod, 1.0, 0.0)
                for f in custom_nb_forces:
                    f.addExclusion(atom1.id, atom2.id)
                if ff.scale_14_vdw == 0:
                    continue
                vdw = ff.get_vdw_term(ff.atom_types[atom1.type], ff.atom_types[atom2.type])
                # We generalize LJ126Term and MieTerm because of minimal computational cost for 1-4 vdW
                if type(vdw) in (LJ126Term, MieTerm):
                    cbforce = pair14_forces.get(MieTerm)
                    if cbforce is None:
                        cbforce = mm.CustomBondForce('C*epsilon*((sigma/r)^n-(sigma/r)^m);'
                                                     'C=n/(n-m)*(n/m)^(m/(n-m))')
                        cbforce.addPerBondParameter('epsilon')
                        cbforce.addPerBondParameter('sigma')
                        cbforce.addPerBondParameter('n')
                        cbforce.addPerBondParameter('m')
                        cbforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
                        cbforce.setForceGroup(ForceGroup.VDW)
                        cbforce.setName('vdW')
                        omm_system.addForce(cbforce)
                        pair14_forces[MieTerm] = cbforce
                    epsilon = vdw.epsilon * ff.scale_14_vdw
                    if type(vdw) == LJ126Term:
                        cbforce.addBond(atom1.id, atom2.id, [epsilon, vdw.sigma, 12, 6])
                    elif type(vdw) == MieTerm:
                        cbforce.addBond(atom1.id, atom2.id,
                                        [epsilon, vdw.sigma, vdw.repulsion, vdw.attraction])
                else:
                    raise Exception('1-4 scaling for vdW terms other than LJ126Term and MieTerm '
                                    'haven\'t been implemented')

        ### Set up Drude particles ##############################################################
        for polar_class in system.polarizable_classes:
            if polar_class == DrudeTerm:
                logger.debug('Setting up Drude polarizations...')
                pforce = mm.DrudeForce()
                pforce.setForceGroup(ForceGroup.DRUDE)
                pforce.setName('Drude')
                omm_system.addForce(pforce)
                parent_idx_thole = {}  # {parent: (index in DrudeForce, thole)} for addScreenPair
                for parent, drude in system.drude_pairs.items():
                    pterm = system.polarizable_terms[parent]
                    n_H = len([atom for atom in parent.bond_partners if atom.symbol == 'H'])
                    alpha = pterm.alpha + n_H * pterm.merge_alpha_H
                    idx = pforce.addParticle(drude.id, parent.id, -1, -1, -1,
                                             drude.charge, alpha, 0, 0)
                    parent_idx_thole[parent] = (idx, pterm.thole)

                # exclude the non-boned interactions between Drude and parent
                # and those concerning Drude particles in 1-2 and 1-3 pairs
                # pairs formed by real atoms have already been handled above
                # also apply thole screening between 1-2 and 1-3 Drude dipole pairs
                drude_exclusions = list(system.drude_pairs.items())
                for atom1, atom2 in pair12 + pair13:
                    drude1 = system.drude_pairs.get(atom1)
                    drude2 = system.drude_pairs.get(atom2)
                    if drude1 is not None:
                        drude_exclusions.append((drude1, atom2))
                    if drude2 is not None:
                        drude_exclusions.append((atom1, drude2))
                    if drude1 is not None and drude2 is not None:
                        drude_exclusions.append((drude1, drude2))
                        idx1, thole1 = parent_idx_thole[atom1]
                        idx2, thole2 = parent_idx_thole[atom2]
                        pforce.addScreenedPair(idx1, idx2, (thole1 + thole2) / 2)
                for a1, a2 in drude_exclusions:
                    nbforce.addException(a1.id, a2.id, 0, 1.0, 0)
                    for f in custom_nb_forces:
                        f.addExclusion(a1.id, a2.id)

                # scale the non-boned interactions concerning Drude particles in 1-4 pairs
                # pairs formed by real atoms have already been handled above
                drude_exceptions14 = []
                for atom1, atom2 in pair14:
                    drude1 = system.drude_pairs.get(atom1)
                    drude2 = system.drude_pairs.get(atom2)
                    if drude1 is not None:
                        drude_exceptions14.append((drude1, atom2))
                    if drude2 is not None:
                        drude_exceptions14.append((atom1, drude2))
                    if drude1 is not None and drude2 is not None:
                        drude_exceptions14.append((drude1, drude2))
                for a1, a2 in drude_exceptions14:
                    charge_prod = a1.charge * a2.charge * ff.scale_14_coulomb
                    nbforce.addException(a1.id, a2.id, charge_prod, 1.0, 0.0)
                    for f in custom_nb_forces:
                        f.addExclusion(a1.id, a2.id)
            else:
                raise Exception('Polarizable terms other that DrudeTerm haven\'t been implemented')

        ### Set up virtual sites ################################################################
        if top.has_virtual_site:
            logger.debug('Setting up virtual sites...')
            for atom in top.atoms:
                vsite = atom.virtual_site
                if type(vsite) == TIP4PSite:
                    O, H1, H2 = vsite.parents
                    coeffs = system.get_TIP4P_linear_coeffs(atom)
                    omm_vsite = mm.ThreeParticleAverageSite(O.id, H1.id, H2.id, *coeffs)
                    omm_system.setVirtualSite(atom.id, omm_vsite)
                elif vsite is not None:
                    raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')

            # exclude the non-boned interactions between virtual sites and parents
            # and particles (atoms, drude particles, virtual sites) in 1-2 and 1-3 pairs
            # TODO Assume no more than one virtual site is attached to each atom
            vsite_exclusions = list(system.vsite_pairs.items())
            for atom, vsite in system.vsite_pairs.items():
                drude = system.drude_pairs.get(atom)
                if drude is not None:
                    vsite_exclusions.append((vsite, drude))
            for atom1, atom2 in pair12 + pair13:
                vsite1 = system.vsite_pairs.get(atom1)
                vsite2 = system.vsite_pairs.get(atom2)
                drude1 = system.drude_pairs.get(atom1)
                drude2 = system.drude_pairs.get(atom2)
                if vsite1 is not None:
                    vsite_exclusions.append((vsite1, atom2))
                    if drude2 is not None:
                        vsite_exclusions.append((vsite1, drude2))
                if vsite2 is not None:
                    vsite_exclusions.append((vsite2, atom1))
                    if drude1 is not None:
                        vsite_exclusions.append((vsite2, drude1))
                if None not in [vsite1, vsite2]:
                    vsite_exclusions.append((vsite1, vsite2))
            for a1, a2 in vsite_exclusions:
                nbforce.addException(a1.id, a2.id, 0, 1.0, 0)
                for f in custom_nb_forces:
                    f.addExclusion(a1.id, a2.id)

            # scale the non-boned interactions between virtual sites and particles in 1-4 pairs
            # TODO Assume no 1-4 LJ interactions on virtual sites
            vsite_exceptions14 = []
            for atom1, atom2 in pair14:
                vsite1 = system.vsite_pairs.get(atom1)
                vsite2 = system.vsite_pairs.get(atom2)
                drude1 = system.drude_pairs.get(atom1)
                drude2 = system.drude_pairs.get(atom2)
                if vsite1 is not None:
                    vsite_exceptions14.append((vsite1, atom2))
                    if drude2 is not None:
                        vsite_exceptions14.append((vsite1, drude2))
                if vsite2 is not None:
                    vsite_exceptions14.append((vsite2, atom1))
                    if drude1 is not None:
                        vsite_exceptions14.append((vsite2, drude1))
                if None not in [vsite1, vsite2]:
                    vsite_exceptions14.append((vsite1, vsite2))
            for a1, a2 in vsite_exceptions14:
                charge_prod = a1.charge * a2.charge * ff.scale_14_coulomb
                nbforce.addException(a1.id, a2.id, charge_prod, 1.0, 0.0)
                for f in custom_nb_forces:
                    f.addExclusion(a1.id, a2.id)

    @staticmethod
    def _setup_nonbonded_as_bonded(system, omm_system):
        if system.polarizable_classes != set():
            raise Exception('Polarizable terms not supported')

        if system.vsite_types != set():
            raise Exception('Virtual sites not supported')

        if system.use_pbc:
            raise Exception('PBC not supported')

        top = system.topology
        ff = system.ff

        logger.debug('Setting up Coulomb interactions...')
        qqforce = mm.CustomBondForce(f'{ONE_4PI_EPS0}*qq/r')
        qqforce.addPerBondParameter('qq')
        qqforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
        qqforce.setForceGroup(ForceGroup.COULOMB)
        qqforce.setName('Coulomb')
        omm_system.addForce(qqforce)

        logger.debug('Setting up vdW interactions...')
        ljforce = mm.CustomBondForce('C*epsilon*((sigma/r)^n-(sigma/r)^m);'
                                     'C=n/(n-m)*(n/m)^(m/(n-m))')
        ljforce.addPerBondParameter('epsilon')
        ljforce.addPerBondParameter('sigma')
        ljforce.addPerBondParameter('n')
        ljforce.addPerBondParameter('m')
        ljforce.setUsesPeriodicBoundaryConditions(system.use_pbc)
        ljforce.setForceGroup(ForceGroup.VDW)
        ljforce.setName('vdW')
        omm_system.addForce(ljforce)

        for mol in top.molecules:
            pairs12, pairs13, pairs14 = mol.get_12_13_14_pairs()
            pairs1n = list(set(itertools.combinations(mol.atoms, 2)) - set(pairs12).union(pairs13).union(pairs14))
            for i, (atom1, atom2) in enumerate(pairs14 + pairs1n):
                vdw = ff.get_vdw_term(ff.atom_types[atom1.type], ff.atom_types[atom2.type])
                epsilon = vdw.epsilon
                qq = atom1.charge * atom2.charge
                if i < len(pairs14):
                    epsilon *= ff.scale_14_vdw
                    qq *= ff.scale_14_coulomb
                if type(vdw) == LJ126Term:
                    ljforce.addBond(atom1.id, atom2.id, [epsilon, vdw.sigma, 12, 6])
                elif type(vdw) == MieTerm:
                    ljforce.addBond(atom1.id, atom2.id, [epsilon, vdw.sigma, vdw.repulsion, vdw.attraction])
                qqforce.addBond(atom1.id, atom2.id, [qq])



import numpy as np


class Atom():
    '''
    An atom is a particle in simulation.

    It can be an real atom, a Drude particle, a virtual site or a coarse-grained bead.

    Parameters
    ----------
    name : str

    Attributes
    ----------
    id : int
        Index of this atom in topology. -1 means information haven\'t been updated by topology
    id_in_mol : int
        Index of this atom in molecule. -1 means information haven\'t been updated by topology
    name : str
        Name of this atom, not necessarily unique
    type : str
        Atom type in force field. Required for assigning force field parameters
    symbol : str
        Atomic symbol. Mainly used for output topology or trajectory
    mass : float
        Mass assigned. Required for output simulation files
    charge : float
        Charge assigned. Required for output simulation files
    alpha : float
        Isotropic polarizability. Required for output Drude polarizable simulation files
        This property is set for parent atom, not Drude particle
    thole : float
        Thole screening for induced dipole. Required for output Drude polarizable simulation files
        This property is set for parent atom, not Drude particle
    formal_charge : int
        Formal charge calculated from valence bond theory. Optionally required by typing engine
    is_drude : bool
        Whether or not this is a Drude particle for polarizable model
    virtual_site : None or subclass of VirtualSite
        None if this atom is not a virtual site.
        Otherwise the instance of subclass of VirtualSite
    has_position : bool
        Whether or not the position of this atom is set
    '''

    def __init__(self, name='UNK'):
        self.id = -1
        self.id_in_mol = -1
        self.name = name
        self.type = 'UNK'
        self.symbol = 'UNK'
        self.mass = 0.
        self.charge = 0.
        self.alpha = 0.  # this property is set for parent atom, not Drude particle
        self.thole = 0.  # this property is set for parent atom, not Drude particle
        self.formal_charge = 0
        self.is_drude = False
        self.virtual_site = None
        self.has_position = False

        self._position = np.zeros(3, dtype=float)
        self._molecule = None
        self._bonds = []
        self._residue = None

    def __repr__(self):
        return f'<Atom: {self.name} {self.id} {self.type}>'

    def __lt__(self, other):
        return self.id < other.id

    def __gt__(self, other):
        return self.id > other.id

    def __deepcopy__(self, memodict={}):
        '''
        id, id_in_mol, virtual_site, residue are not copied, because these information depends on other atoms.
        '''
        atom = Atom()
        atom.name = self.name
        atom.type = self.type
        atom.symbol = self.symbol
        atom.mass = self.mass
        atom.charge = self.charge
        atom.alpha = self.alpha
        atom.thole = self.thole
        atom.formal_charge = self.formal_charge
        atom.is_drude = self.is_drude
        atom.has_position = self.has_position
        atom._position[:] = self._position[:]
        return atom

    @property
    def molecule(self):
        '''
        The molecule this atom belongs to

        Returns
        -------
        molecule : Molecule
        '''
        return self._molecule

    @property
    def residue(self):
        '''
        The residue this atom belongs to

        Returns
        -------
        residue : Residue
        '''
        return self._residue

    @property
    def bonds(self):
        '''
        All the bonds involving this atom.

        Bonds of Drude dipoles are included if exist.

        Returns
        -------
        bonds : list of Bond
        '''
        return self._bonds

    @property
    def bond_partners(self):
        '''
        All the bond partners of this atom.

        Drude particles are included if exist.

        Returns
        -------
        partners : list of Atom
        '''
        return [bond.atom2 if bond.atom1 == self else bond.atom1 for bond in self._bonds]

    @property
    def position(self):
        '''
        The position of this atom

        :setter: Set the position of this atom

        Returns
        -------
        position : array_like
            The position is a numpy array of shape (3,)
        '''
        return self._position

    @position.setter
    def position(self, value):
        if not isinstance(value, (list, tuple, np.ndarray)) or len(value) != 3:
            raise ValueError('position should has three elements')
        self._position[:] = value
        self.has_position = True

    @property
    def rdatom(self):
        '''
        The `rdkit.Chem.Atom` object associated with this atom.

        It exists only if this atom belongs to a molecule.

        Returns
        -------
        rdbond : rdkit.Chem.Atom
        '''
        if self.molecule is None:
            return None

        return self.molecule.rdmol.GetAtomWithIdx(self.id_in_mol)



from .frame import Frame
from .handler import TrjHandler


class Trajectory():
    '''
    A Trajectory is made of a series of Frames.

    It is assumed that all frames have the same number of atoms.
    If it's not the case (e.g. the LammpsDump can have variable number of atoms in each frame),
    then this class may not be able to work correctly.

    The file(s) can be opened in 'r', 'w' or 'a' modes, which mean read, write and append.
    The frames are loaded on request for performance issue,
    therefore a file handler is kept open until :func:`close` is called.
    The handler class used for parsing the trajectory is determined by the extension of the file.
    It can also be specified explicitly if non-standard extension is used for the name of trajectory file.

    Several files can be opened at the same time (even for files of different format),
    which is useful for processing truncated trajectories.
    In this case, the trajectory is read-only.

    Parameters
    ----------
    file : str or list of str
        If it's a str, then open the file with corresponding parser.
        if it's a list of str, then call :class:`CombinedTrj` to open all of the files.
    mode : ['r', 'w', 'a']
    Handler : subclass of TrjHandler
        Specify the handler class for parsing the trajectory.
        If set to None, then the handler class will be determined from the extension of the file name.

    Attributes
    ----------
    n_atom : int
        Number of atoms in the trajectory. -1 means unknown yet
    n_frame : int
        Number of frames in the trajectory. -1 means unknown yet
    frame : Frame or None
        The latest frame been read by :func:`read_frame`.
        None means no frame has been read.

    Examples
    --------
    >>> trj = Trajectory('input.dcd')
    >>> frame = trj.read_frame(0)
    >>> frames = trj.read_frames([1,2,5])
    >>> trj.close()

    >>> trj = Trajectory('output.dcd', mode='a')
    >>> trj.write_frame(frame)
    >>> trj.close()

    >>> trj = Trajectory(['input1.dcd', 'input2.xtc'])
    '''

    def __init__(self, file, mode='r', Handler=None):
        modes_allowed = ('r', 'w', 'a')
        if mode not in modes_allowed:
            raise Exception('mode should be one of %s' % str(modes_allowed))

        if Handler is None:
            try:
                Handler = TrjHandler.get_handler_for_file(file)
            except:
                raise Exception('Cannot determine the format of the trajectory file. Try to specify the handler class')

        self._handler: TrjHandler = Handler(file, mode)
        self._opened: bool = True
        self._mode: str = mode
        self.frame: Frame = None

        if mode == 'r':
            self.n_atom, self.n_frame = self._handler.get_info()

    def __del__(self):
        self.close()

    def __repr__(self):
        return f'<Trajectory: {self.n_frame} frames {self.n_atom} atoms>'

    def close(self):
        '''
        Close the opened trajectory file(s).
        '''
        try:
            self._handler.close()
        except:
            pass
        self._opened = False

    def read_frame(self, i_frame: int):
        '''
        Read a single frame from the trajectory.

        For the best of performance, a cached frame is used.
        When this method is called, the position and cell etc in the cached frame are updated,
        and the reference of this frame will be returned.
        Therefore, if you want to handle several different frames at the same time,
        this method should not be used because all the frames actually point to the same frame.
        In this case, use :func:`read_frames` instead.

        i_frame should be in the range of [-1, n_frame), otherwise and Exception will be raised.
        -1 means the last frame.

        Parameters
        ----------
        i_frame : int
            The index of the frame in trajectory.

        Returns
        -------
        frame : Frame

        '''
        if self._mode != 'r' or not self._opened:
            raise Exception('mode != "r" or closed trajectory')

        if i_frame >= self.n_frame:
            raise Exception('i_frame should be smaller than %i' % self.n_frame)

        if self.frame is None:
            if self.n_atom == -1:
                raise Exception('Invalid number of atoms')
            self.frame = Frame(self.n_atom)

        # Reset the information in self.frame in case the frames read from different trajectory files pollute each other for CombinedTrajectory
        self.frame.reset()

        if i_frame == -1:
            i_frame = self.n_frame - 1
        self._handler.read_frame(i_frame, self.frame)
        return self.frame

    def read_frames(self, i_frames: [int]):
        '''
        Read a bunch of frames from the trajectory.

        Unlike :func:`read_frame`, the cached frame is not used.
        Instead, a new Frame object is constructed for each frame.
        This method should be called when you want to multiprocess several frames in parallel.

        All the items in i_frames should be in the range of [-1, n_frame), otherwise and Exception will be raised.
        -1 means the last frame.

        Parameters
        ----------
        i_frames : list of int

        Returns
        -------
        frames : list of Frame

        '''
        if self._mode != 'r' or not self._opened:
            raise Exception('mode != "r" or closed trajectory')

        if any(i >= self.n_frame for i in i_frames):
            raise Exception('i_frame should be smaller than %i' % self.n_frame)

        frames = [Frame(self.n_atom) for _ in i_frames]
        for ii, i_frame in enumerate(i_frames):
            if i_frame == -1:
                i_frame = self.n_frame - 1
            self._handler.read_frame(i_frame, frames[ii])
        return frames

    def write_frame(self, frame, topology=None, subset=None, **kwargs):
        '''
        Write one frame to the opened trajectory file.

        It requires the trajectory opened in 'w' or 'a' mode.
        Some trajectory format (e.g. GRO, XYZ) record part of the topology information,
        therefore, a topology should be provided for these formats.

        It is possible to write only a subset of atoms in the frame.
        In this case, a list of indexes of these atoms should be provided.
        Make sure the number of atoms in the subset is the same for every frame been written.

        Parameters
        ----------
        frame : Frame
        topology : Topology, optional
        subset : list of int, optional
        kwargs : dict
        '''
        if self._mode not in ('w', 'a') or not self._opened:
            raise Exception('mode not in ("w", "a") or closed trajectory')

        self._handler.write_frame(frame, topology=topology, subset=subset, **kwargs)

    @staticmethod
    def open(file, mode='r'):
        '''
        Open (a) trajectory file(s). Deprecated. Use the constructor instead.

        Parameters
        ----------
        file : str or list of str
        mode : ['r', 'w', 'a']
        '''

        return Trajectory(file, mode)

    @staticmethod
    def read_frame_from_file(file, i_frame=0):
        '''
        Read one single frame from a trajectory file.

        This is simply a shortcut for

        >>> trj = Trajectory('input.dcd')
        >>> frame = trj.read_frame(i_frame)
        >>> trj.close()

        By default, the first frame is read.
        Otherwise, argument `i_frame` should be set.
        If i_frame set to -1, then the last frame is read.

        Parameters
        ----------
        file : str
        i_frame : int

        Returns
        -------
        frame : Frame

        '''
        trj = Trajectory(file, 'r')
        frame = trj.read_frame(i_frame)
        trj.close()
        return frame



from .checkpointreporter import CheckpointReporter
from .drudetemperaturereporter import DrudeTemperatureReporter
from .groreporter import GroReporter
from .statedatareporter import StateDataReporter
from .viscosityreporter import ViscosityReporter


from mstk.errors import RDKitError

try:
    from rdkit.Chem import AllChem as Chem
except ImportError:
    RDKIT_FOUND = False
else:
    RDKIT_FOUND = True


def create_mol_from_smiles(smiles):
    '''
    Create a RDKit molecule object from SMILES string.

    Parameters
    ----------
    smiles : str

    Returns
    -------
    rdmol : rdkit.rdchem.Mol
    '''
    if not RDKIT_FOUND:
        raise ImportError('RDKit is required for parsing SMILES')

    try:
        rdmol = Chem.MolFromSmiles(smiles)
    except:
        raise RDKitError('Invalid SMILES')

    rdmol = Chem.AddHs(rdmol)
    if Chem.EmbedMolecule(rdmol, useRandomCoords=True) == -1:
        if Chem.EmbedMolecule(rdmol) == -1:
            raise RDKitError('Failed generating 3D coordinates')

    return rdmol



import os
import random
import subprocess
import sys
from subprocess import PIPE

from mstk.errors import PackmolError


class Packmol:
    '''
    Wrapper for Packmol for building initial coordinates for simulation

    Parameters
    ----------
    packmol_bin : str
        Path of the packmol binary executable
    '''

    def __init__(self, packmol_bin):
        self.PACKMOL_BIN = packmol_bin

    def build_box(self, files, numbers, output, size=None, length=None, slab=None, slab_multiple=False,
                  tolerance=0.2, seed=None, inp_file='build.inp', silent=False):
        '''
        Build a box directly from molecule files.

        Currently, rectangular box of following models are supported:

        * Homogeneous bulk liquid or gas model
        * Liquid-vapor interface model
        * Multi-component liquid-liquid interface model

        The input molecule files can be in XYZ or PDB format.
        The format for all input files and output files should be the same.

        Parameters
        ----------
        files : list of str
            List of input XYZ or PDB files
        numbers : list of int
            Numbers to be packed of each molecules
        output : str
            File name of the ouput XYZ or PDB file
        size : list of float, or None
            Size of the packed rectangular box, in unit of nanometer
            If not set, argument `length` should be provided.
        length : float or None
            Length of the packed cubic box, in unit of nanometer
            If not set, argument `size` should be provided.
        slab : float or None
            If a vapor-liquid interface model is to be packed, it gives the z-coordinates of the interface, in unit of nanometer
            It conflicts with argument `slab_multiple`.
        slab_multiple : bool or None
            Set it to True if a liquid-liquid interface model is to be packed.
            It conflicts with argument `slab`.
        tolerance : float
            The minimum distance between atoms belongs to different molecules, in unit of nanometer
        seed : int or None
            The seed for randomizing positions.
            If not provided, a randomly generated seed will be used.
        inp_file : str
            The input file for running packmol to be written.
        silent : bool
            If set to True, packmol will run silently without writing output on the screen.
            The error message will still be written on the screen.
        '''
        Packmol.gen_inp(files, numbers, output, size, length, slab, slab_multiple, tolerance, seed, inp_file)

        # TODO subprocess PIPE not work for Packmol new version, do not know why
        if silent:
            if os.name == 'nt':
                os.system(self.PACKMOL_BIN + ' < %s > nul' % inp_file)
            else:
                os.system(self.PACKMOL_BIN + ' < %s > /dev/null' % inp_file)
        else:
            os.system(self.PACKMOL_BIN + ' < %s' % inp_file)

            # (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
            # sp = subprocess.Popen([self.PACKMOL_BIN], stdin=PIPE, stdout=stdout, stderr=stderr)
            # sp.communicate(input=inp.encode())

    @staticmethod
    def gen_inp(files, numbers, output, size=None, length=None, slab=None, slab_multiple=False,
                tolerance=0.2, seed=None, inp_file='build.inp'):
        '''
        Generate input file for running packmol.

        It is the same as :func:`build_box` except that packmol is not called after input file generated.
        See :func:`build_box` for the details of parameters.
        '''
        if len(files) == 0:
            raise PackmolError('No files provided')
        if len(files) != len(numbers):
            raise PackmolError('Invalid numbers')

        extensions = {filename.split('.')[-1].lower() for filename in files}
        if len(extensions) > 1:
            raise PackmolError('All file types should be the same')
        filetype = extensions.pop()

        if size is not None:
            if len(size) != 3:
                raise PackmolError('Invalid box size')
            else:
                size = size[:]
        elif length is not None:
            size = [length, length, length]
        else:
            raise PackmolError('Box size needed')

        seed = seed or random.randint(1e7, 1e8)

        inp = (
            'filetype {filetype}\n'
            'tolerance {tolerance}\n'
            'output {output}\n'
            'seed {seed}\n'.format(filetype=filetype, tolerance=tolerance * 10, output=output, seed=seed)
        )

        # liquid-gas interface
        if slab is not None:
            box_liq = '0 0 0 %f %f %f' % (size[0] * 10, size[1] * 10, slab * 10)  # nm -> A
            box_gas = '0 0 %f %f %f %f' % (slab * 10, size[0] * 10, size[1] * 10, size[2] * 10)  # nm -> A
            for i, filename in enumerate(files):
                # put 1/50 molecules in gas phase. Do not put too many in case of nucleation in gas phase
                if numbers[i] == 0:
                    continue
                n_gas = numbers[i] // 50
                n_liq = numbers[i] - n_gas
                inp += (
                    'structure {filename}\n'
                    'number {n_liq}\n'
                    'inside box {box_liq}\n'
                    'end structure\n'
                    'structure {filename}\n'
                    'number {n_gas}\n'
                    'inside box {box_gas}\n'
                    'end structure\n'.format(filename=filename, n_liq=n_liq, n_gas=n_gas,
                                             box_liq=box_liq, box_gas=box_gas)
                )

        else:
            for i, filename in enumerate(files):
                if numbers[i] == 0:
                    continue
                # slab model for multiple components
                if slab_multiple:
                    lz_per_slab = size[3] / len(numbers)
                    box = '0 0 %f %f %f %f' % (
                        i * lz_per_slab * 10, size[0] * 10, size[1] * 10, (i + 1) * lz_per_slab * 10)
                else:
                    box = '0 0 0 %f %f %f' % tuple(element * 10 for element in size)

                inp += (
                    'structure {filename}\n'
                    'number {number}\n'
                    'inside box {box}\n'
                    'end structure\n'.format(filename=filename, number=numbers[i], box=box)
                )

        with open(inp_file, 'w') as f:
            f.write(inp)



class Typer:
    '''
    Base class for typing engines.

    Typing engine assigns atom types for atoms in a molecule or topology by some predefined rules.
    It is the very first step for force field assignment.
    It is also the basis of force field development.
    A well defined typing rule will make the force field development much less painful.
    '''

    def type(self, top_or_mol):
        '''
        Assign types for all atoms in a topology or molecule.

        The :attr:`~mstk.topology.Atom.type` attribute of all atoms in the topology/molecule will be updated.

        Parameters
        ----------
        top_or_mol: Topology, Molecule
        '''
        from mstk.topology import Topology, Molecule

        if type(top_or_mol) is Topology:
            for mol in top_or_mol.molecules:
                self._type_molecule(mol)
        elif type(top_or_mol) is Molecule:
            self._type_molecule(top_or_mol)
        else:
            raise Exception('A topology or molecule is expected')

    def _type_molecule(self, molecule):
        '''
        Assign types for all the atoms in the molecule.
        This method should be implemented by subclasses.

        Parameters
        ----------
        molecule : Molecule
        '''
        raise NotImplementedError('This method haven\'t been implemented')



#!/usr/bin/env python3

import sys
import argparse
import numpy as np

from mstk.topology import Topology
from mstk.trajectory import Trajectory
from mstk import logger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('input', nargs='+', type=str, help='trajectory files')
    parser.add_argument('-p', '--top', required=True, type=str, help='topology file')
    parser.add_argument('-o', '--output', required=True, type=str, help='output trajectory file')
    parser.add_argument('-b', '--begin', default=0, type=int,
                        help='read from this frame (included). '
                             'A negative value has the same meaning as in Python list')
    parser.add_argument('-e', '--end', type=int,
                        help='read until this frame (not included). '
                             'A negative value has the same meaning as in Python list. '
                             'If not set, will read until the end of the trajectory')
    parser.add_argument('--skip', default=1, type=int, help='read every N frames')
    parser.add_argument('--ignore', nargs='+', default=[], type=str, help='ignore these molecule names')
    parser.add_argument('--ignoreatom', nargs='+', default=[], type=str, help='ignore these atom types')
    parser.add_argument('--wrapfirst', action='store_true',
                        help='shift the positions of all atoms so that all residues in the first frame are in the main cell')
    parser.add_argument('--box', nargs=3, default=[-1, -1, -1], type=float,
                        help='overwrite the box dimensions')
    parser.add_argument('--shift', nargs=3, default=[0, 0, 0], type=float,
                        help='shift the positions of all atoms')
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()

    top = Topology.open(args.top)
    logger.info(top)

    trj = Trajectory.open(args.input)
    logger.info(trj)

    if (top.n_atom != trj.n_atom):
        raise Exception('Number of atoms in topology and trajectory files do not match')

    trj_out = Trajectory.open(args.output, 'w')

    if args.ignore or args.ignoreatom:
        subset = [atom.id for atom in top.atoms
                  if atom.type not in args.ignoreatom and atom.molecule.name not in args.ignore]
    else:
        subset = None

    if args.begin < 0:
        args.begin += trj.n_frame
    if args.end is None or args.end > trj.n_frame:
        args.end = trj.n_frame
    elif args.end < 0:
        args.end += trj.n_frame

    cell_offset = np.zeros((top.n_atom, 3), dtype=int)
    if args.wrapfirst:
        frame = trj.read_frame(args.begin)
        for res in top.residues:
            ids = [atom.id for atom in res.atoms]
            positions = frame.positions[ids]
            center = np.sum(positions, axis=0) / len(positions)
            cell_offset[ids] = np.floor(center / frame.cell.size).astype(int)

    if any(val != 0 for val in args.shift):
        pos_shift = np.array([args.shift] * top.n_atom)
    else:
        pos_shift = None

    for i in range(args.begin, args.end, args.skip):
        sys.stdout.write('\r    %i' % i)
        frame = trj.read_frame(i)
        box = np.array([args.box[k] if args.box[k] != -1 else frame.cell.size[k] for k in range(3)])
        frame.cell.set_box(box)
        if args.wrapfirst:
            frame.positions -= cell_offset * box
        if pos_shift is not None:
            frame.positions += pos_shift
        trj_out.write_frame(frame, top, subset=subset)
    trj_out.close()



'''
Pre-defined physical constants.
'''

PI = 3.1415926535
PI_SQRT = PI ** 0.5
VACUUM_PERMITTIVITY = 8.854_187_812_8E-12  # farad/meter
ELEMENTARY_CHARGE = 1.602_176_62E-19  # coulomb
AVOGADRO = 6.022_140_76E23
CENTI = 1E-2
MILLI = 1E-3
MICRO = 1E-6
NANO = 1E-9
PICO = 1E-12
# q^2/nm -> 138.93545522028575 kJ/mol
ONE_4PI_EPS0 = 1 / (4 * PI * VACUUM_PERMITTIVITY) \
               * ELEMENTARY_CHARGE ** 2 / NANO / 1000 * AVOGADRO
RAD2DEG = 180.0 / PI
DEG2RAD = PI / 180.0



from mstk.chem.constant import *
from mstk.chem.element import Element
from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology
from mstk import logger


class Pdb:
    '''
    Generate Topology from PDB file.

    Atom name, residue information, unit cell and positions are parsed.
    The atom name colume will be used as both atom name and atom type.
    The connectivity is parsed if CONECT section is provided in the PDB file.

    Only the first frame will be considered.

    Parameters
    ----------
    file : str
    split_molecule : str
        Can be 'residue', 'whole', 'bond' or 'auto'. Default is 'auto'.
        If set to 'residue', each residue will be parsed as a molecule. It works only if there is no inter-residue bonds.
        If set to 'whole', all the atoms will be put into one molecule.
        If set to 'bond', the atoms will be grouped into molecules based on connectivity.
        if set to 'auto', it will try 'residue' first. If there's inter-residue connectivity, will use 'whole' instead.

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> pdb = Pdb('input.pdb')
    >>> topology = pdb.topology

    >>> Pdb.save_to(topology, 'output.pdb')
    '''

    def __init__(self, file, **kwargs):
        self.topology = Topology()
        self._molecule = Molecule()
        self._parse(file, **kwargs)

    def _parse(self, file, split_molecule='auto'):
        with open(file) as f:
            lines = f.read().splitlines()

        mol = self._molecule
        atom_ids = {}  # {int: Atom}
        res_names = {}  # {int: str}
        res_atoms = {}  # {int: [Atom]}
        _atom_started = False
        for line in lines:
            line = line.rstrip()
            if _atom_started and line == 'ENDMDL':
                break
            if line.startswith('CRYST1'):
                a, b, c, alpha, beta, gamma = list(map(float, line.split()[1:7]))
                self.topology.cell.set_box([[a / 10, b / 10, c / 10],
                                            [alpha * DEG2RAD, beta * DEG2RAD, gamma * DEG2RAD]])
            if line.startswith('ATOM') or line.startswith('HETATM'):
                _atom_started = True
                atom_id = int(line[6:11])
                atom_name = line[12:16].strip()
                res_name = line[17:21].strip()
                res_id = int(line[22:26])
                x = float(line[30:38]) / 10
                y = float(line[38:46]) / 10
                z = float(line[46:54]) / 10
                if len(line) > 76:
                    symbol = line[76:78].strip()
                    element = Element(symbol)
                else:
                    element = Element.guess_from_atom_type(atom_name)

                atom = Atom(atom_name)
                atom.type = atom_name
                atom.position = (x, y, z,)
                atom.symbol = element.symbol
                atom.mass = element.mass
                mol.add_atom(atom)

                atom_ids[atom_id] = atom
                if res_id not in res_names:
                    res_names[res_id] = res_name
                    res_atoms[res_id] = []
                res_atoms[res_id].append(atom)

        for resid, resname in res_names.items():
            atoms = res_atoms[resid]
            mol.add_residue(resname, atoms, refresh_residues=False)
        mol.refresh_residues()

        prev_section = ''
        connects = {}
        for line in lines:
            line = line.rstrip()
            section = line[:6]
            # only read the first frame
            if prev_section == 'CONECT' and section != 'CONECT':
                break
            if section == 'CONECT':
                id1 = int(line[6:11])
                id2 = int(line[11:16])
                connects[tuple(sorted([id1, id2]))] = None
                if len(line) > 16:
                    id3 = int(line[16:21])
                    connects[tuple(sorted([id1, id3]))] = None
                if len(line) > 21:
                    id4 = int(line[21:26])
                    connects[tuple(sorted([id1, id4]))] = None
                if len(line) > 26:
                    id5 = int(line[26:31])
                    connects[tuple(sorted([id1, id5]))] = None
            prev_section = section

        for id1, id2 in connects:
            atom1 = atom_ids[id1]
            atom2 = atom_ids[id2]
            mol.add_bond(atom1, atom2)

        if split_molecule == 'residue':
            self.topology.update_molecules(mol.split_residues())
        elif split_molecule == 'whole':
            self.topology.update_molecules([mol])
        elif split_molecule == 'bond':
            self.topology.update_molecules(mol.split(consecutive=True))
        elif split_molecule == 'auto':
            try:
                pieces = self._molecule.split_residues()
            except:
                self.topology.update_molecules([mol])
            else:
                self.topology.update_molecules(pieces)
        else:
            raise Exception(f'Invalid option for split_molecule: {split_molecule}')
        self.topology.generate_angle_dihedral_improper()

    @staticmethod
    def save_to(top, file, **kwargs):
        '''
        Save topology into a PDB file

        Parameters
        ----------
        top : Topology
        file : str
        '''
        if not top.has_position:
            raise Exception('Position is required for writing PDB file')

        if top.n_atom > 99999:
            logger.warning('Number of atoms exceeds the maximum supported by PDB (99999)')

        def format_float(value, max_length, default_decimal):
            if value >= 10 ** max_length or value <= -10 ** (max_length - 1):
                raise Exception(f'{value} too long for formatter')
            formatter = f'%{max_length}.{default_decimal}f'
            return (formatter % value)[:max_length]

        lengths = top.cell.lengths * 10
        angles = top.cell.angles
        string = 'REMARK   Created by mstk\n'
        string += 'CRYST1%9s%9s%9s%7s%7s%7s P 1           1 \n' % (
            *(format_float(x, 9, 3) for x in lengths),
            *(format_float(x * RAD2DEG, 7, 2) for x in angles))

        for atom in top.atoms:
            pos = atom.position * 10  # convert from nm to A
            resname = atom.residue.name
            resid = atom.residue.id + 1
            line = 'HETATM%5d %4s %-4s %4d    %8s%8s%8s                      %2s\n' % (
                (atom.id + 1) % 100000, atom.name[:4], resname[:4], resid % 10000,
                *[format_float(x, 8, 3) for x in pos], atom.symbol[:2])
            string += line

        for atom in top.atoms:
            partners = [a for a in atom.bond_partners if a.id > atom.id]
            if len(partners) > 0:
                line = 'CONECT%5d' % ((atom.id + 1) % 100000)
                for i, partner in enumerate(partners):
                    if i > 0 and i % 4 == 0:
                        line += '\nCONECT%5d' % ((atom.id + 1) % 100000)
                    line += '%5d' % ((partner.id + 1) % 100000)
                line += '\n'
                string += line

        with open(file, 'wb') as f:
            f.write(string.encode())


Topology.register_format('.pdb', Pdb)



import abc


class DocstringMeta(abc.ABCMeta):
    ''' Metaclass that allows docstring inheritance '''

    def __new__(mcls, classname, bases, cls_dict):
        cls = abc.ABCMeta.__new__(mcls, classname, bases, cls_dict)
        mro = cls.__mro__[1:]
        for name, member in cls_dict.items():
            if not getattr(member, '__doc__'):
                for base in mro:
                    try:
                        member.__doc__ = getattr(base, name).__doc__
                        break
                    except AttributeError:
                        pass
        return cls



import numpy as np
import pandas as pd
from io import StringIO
from mstk import logger
from mstk.trajectory.handler import TrjHandler


class LammpsTrj(TrjHandler):
    '''
    Read step, cell and atomic positions (and charges if provided) from dump file of Lammps.

    Velocities are ignored.
    Because the topology information are detailed in data file, the mol, type, element in dump file are also ignored.
    The cell used by Lammps is not originated at (0, 0, 0) usually.
    This handler will translate the positions of atoms based on the lower bound of the cell.

    Writing to LammpsDump is not supported yet.
    '''

    def __init__(self, trj_file, mode='r'):
        super().__init__()

        if mode == 'r':
            # open it in binary mode so that we can correctly seek despite of line ending
            self._file = open(trj_file, 'rb')
        else:
            raise Exception('Writing support for LammpsTrj haven\'t been implemented')

    def get_info(self):
        try:
            self._file.readline()
            self._file.readline()
            self._file.readline()
            self.n_atom = int(self._file.readline())
        except:
            print('Invalid lammpstrj file')
            raise
        self._file.seek(0)

        # read in the file once and build a list of line offsets
        # the last element is the length of whole file
        self._line_offset = [0]
        offset = 0
        for line in self._file:
            offset += len(line)
            self._line_offset.append(offset)
        self._file.seek(0)

        # build a list of frame offsets
        self.n_frame = len(self._line_offset) // (9 + self.n_atom)
        self._frame_offset = []
        for i in range(self.n_frame + 1):
            line_start = (9 + self.n_atom) * i
            self._frame_offset.append(self._line_offset[line_start])

        return self.n_atom, self.n_frame

    def read_frame(self, i_frame, frame):
        # skip to frame i and read only this frame
        self._file.seek(self._frame_offset[i_frame])
        string = self._file.read(self._frame_offset[i_frame + 1] - self._frame_offset[i_frame]).decode()
        lines = string.splitlines()
        frame.step = int(lines[1])
        try:
            xlo, xhi = tuple(map(lambda x: float(x) / 10, lines[5].split()))  # convert from A to nm
            ylo, yhi = tuple(map(lambda x: float(x) / 10, lines[6].split()))
            zlo, zhi = tuple(map(lambda x: float(x) / 10, lines[7].split()))
            frame.cell.set_box([xhi - xlo, yhi - ylo, zhi - zlo])
        except:
            # read triclinic box, convert from A to nm
            xlo, xhi, bx = tuple(map(lambda x: float(x) / 10, lines[5].split()))
            ylo, yhi, cx = tuple(map(lambda x: float(x) / 10, lines[6].split()))
            zlo, zhi, cy = tuple(map(lambda x: float(x) / 10, lines[7].split()))
            frame.cell.set_box([[xhi - xlo, 0, 0], [bx, yhi - ylo, 0], [cx, cy, zhi - zlo]])

        tokens = lines[8].split()
        title = tokens[2:]

        data = '\n'.join(lines[9:])
        df = pd.read_csv(StringIO(data), header=None, index_col=None, names=title, sep=r'\s+')
        wrapped = False
        if 'x' in df.columns:
            df['x'] = df['x'] / 10  # convert from A to nm
            df['y'] = df['y'] / 10
            df['z'] = df['z'] / 10
            wrapped = True
        elif 'xs' in df.columns:
            df['x'] = df.xs * frame.cell.size[0] + xlo
            df['y'] = df.ys * frame.cell.size[1] + ylo
            df['z'] = df.zs * frame.cell.size[2] + zlo
            wrapped = True
        elif 'xu' in df.columns:
            df['x'] = df.xu / 10  # convert from A to nm
            df['y'] = df.yu / 10
            df['z'] = df.zu / 10
        elif 'xsu' in df.columns:
            df['x'] = df.xsu * frame.cell.size[0] + xlo
            df['y'] = df.ysu * frame.cell.size[1] + ylo
            df['z'] = df.zsu * frame.cell.size[2] + zlo
        if wrapped:
            if 'ix' not in df.columns:
                logger.warning('Image flag not found for wrapped positions')
            else:
                # ix is revered words for pandas, so use df['ix'] instead of df.ix
                df['x'] += df['ix'] * frame.cell.size[0]
                df['y'] += df['iy'] * frame.cell.size[1]
                df['z'] += df['iz'] * frame.cell.size[2]

        frame.has_charge = 'q' in df.columns
        for row in df.itertuples():
            frame.positions[row.id - 1] = np.array([row.x, row.y, row.z])
            if frame.has_charge:
                frame.charges[row.id - 1] = float(row.q)


TrjHandler.register_format('.lammpstrj', LammpsTrj)
TrjHandler.register_format('.ltrj', LammpsTrj)



class Formula:
    '''
    Parse elements and numbers from a non-standardized chemical formula.

    e.g. H4C3(COH2)2 will be parsed as five carbon, eight hydrogen and two oxygen.

    Parameters
    ----------
    mol_str : str

    Attributes
    ----------
    atoms : dict, [str, int]
        Dict of the symbol and number of each chemical element.
    '''

    def __init__(self, mol_str):
        chars = Formula._extract_chars(mol_str)
        chars = Formula._expand_chars(chars)
        atoms = Formula._count_atoms(chars)
        self.atoms = Formula._hill_order(atoms)

    @staticmethod
    def _extract_chars(formula):
        chars = []
        i = 0
        while i < len(formula):
            c = formula[i]
            if c.isupper():
                chars.append(c)
                i += 1
                while i < len(formula):
                    if formula[i].islower():
                        chars[-1] += formula[i]
                        i += 1
                    else:
                        break
            elif c.isdigit():
                chars.append(c)
                i += 1
                while i < len(formula):
                    if formula[i].isdigit():
                        chars[-1] += formula[i]
                        i += 1
                    else:
                        break
            elif c == '(' or c == ')':
                chars.append(c)
                i += 1
            else:
                raise Exception('Invalid character: %s' % c)

        if chars[0].isdigit():
            raise Exception('Invalid formula')
        if chars.count('(') != chars.count(')'):
            raise Exception('Unmatched brackets')
        return chars

    @staticmethod
    def _expand_chars(chars):
        chars = chars[:]
        temp_expanded_chars = []
        temp_chars = []
        while '(' in chars:
            _start = False
            i = 0
            while i < len(chars):
                c = chars[i]
                if c == '(':
                    _start = True
                    temp_expanded_chars = chars[:i]
                    temp_chars = []
                    i += 1
                elif c == ')':
                    if not _start:
                        raise Exception('Unmatched brackets')
                    if i + 1 < len(chars) and chars[i + 1].isdigit():
                        n = int(chars[i + 1])
                        temp_expanded_chars += temp_chars * n
                        if i + 2 < len(chars):
                            temp_expanded_chars += chars[i + 2:]
                    else:
                        n = 1
                        temp_expanded_chars += temp_chars * n
                        if i + 1 < len(chars):
                            temp_expanded_chars += chars[i + 1:]
                    chars = temp_expanded_chars
                    break
                else:
                    if _start:
                        temp_chars.append(c)
                    i += 1

        return chars

    @staticmethod
    def _count_atoms(chars):
        counts = {}
        for i in range(len(chars)):
            c = chars[i]
            if c.isdigit():
                counts[chars[i - 1]] += int(c) - 1
            else:
                if c not in counts:
                    counts[c] = 1
                else:
                    counts[c] += 1
        return counts

    @staticmethod
    def _hill_order(atoms):
        count_C = atoms.pop('C') if 'C' in atoms else 0
        count_H = atoms.pop('H') if 'H' in atoms else 0
        symbols = list(atoms.keys())
        symbols.sort()
        atoms_sorted = {}
        if count_C != 0:
            atoms_sorted['C'] = count_C
        if count_H != 0:
            atoms_sorted['H'] = count_H
        for s in symbols:
            atoms_sorted[s] = atoms[s]
        return atoms_sorted

    def to_str(self):
        '''
        Return the standardized formula in hill order.

        Returns
        -------
        formula_str : str
        '''

        def num2str(num):
            if num == 1:
                return ''
            else:
                return str(num)

        return ''.join([symbol + num2str(num) for symbol, num in self.atoms.items()])

    @property
    def n_heavy(self):
        '''
        Number of heavy atoms (non-hydrogen atoms) in this formula.

        Returns
        -------
        n_heavy : int
        '''
        n = 0
        for k, v in self.atoms.items():
            if k != 'H':
                n += v
        return n

    @property
    def n_h(self):
        '''
        Number of hydrogen atoms in this formula.

        Returns
        -------
        n_h : int
        '''
        n = 0
        for k, v in self.atoms.items():
            if k == 'H':
                n += v
        return n



import itertools
import copy
import numpy as np
from mstk import logger
from mstk.chem.rdkit import create_mol_from_smiles
from mstk.chem.element import Element
from mstk.forcefield.ffterm import *
from .atom import Atom
from .virtualsite import *
from .connectivity import *
from .unitcell import UnitCell
from .residue import Residue
from .geometry import find_clusters, find_clusters_consecutive


class Molecule():
    '''
    A molecule is defined as atoms and the connectivity between them.

    The term `molecule` is not strictly a chemical molecule.
    Some atoms may not be connected to any other atoms in the same molecule.
    However, there can not be bonds connecting atoms belong to different molecules.
    Drude particles and virtual sites are also considered as atoms.
    All bond, angles, dihedrals and impropers should be defined explicitly.

    Parameters
    ----------
    name : str

    Attributes
    ----------
    id : int
        Index of this molecule in topology. -1 means information haven\'t been updated by topology
    name : str
        Name of the molecule, not necessarily unique
    '''

    def __init__(self, name='UNK'):
        self.id = -1
        self._name = name
        self._topology = None
        self._atoms: [Atom] = []
        self._bonds: [Bond] = []
        self._angles: [Angle] = []
        self._dihedrals: [Dihedral] = []
        self._impropers: [Improper] = []
        self._rdmol = None  # this is for typing based on SMARTS
        self._is_rdmol_valid = False
        self._default_residue = Residue(name)
        self._added_residues = []

    def __repr__(self):
        return f'<Molecule: {self.name} {self.id}>'

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val
        self._default_residue.name = val

    def __deepcopy__(self, memodict={}):
        '''
        If there are virtual sites, they will be constructed with new atoms also
        If there are residue information, they will also be constructed

        topology will not be copied.
        id will not be copied, because it relies on other molecules in the topology
        '''
        mol = Molecule(self.name)
        for atom in self._atoms:
            mol.add_atom(copy.deepcopy(atom))
        for bond in self._bonds:
            idx1 = bond.atom1.id_in_mol
            idx2 = bond.atom2.id_in_mol
            mol.add_bond(mol._atoms[idx1], mol._atoms[idx2], bond.order)
        for angle in self._angles:
            idx1 = angle.atom1.id_in_mol
            idx2 = angle.atom2.id_in_mol
            idx3 = angle.atom3.id_in_mol
            mol.add_angle(mol._atoms[idx1], mol._atoms[idx2], mol._atoms[idx3])
        for dihedral in self._dihedrals:
            idx1 = dihedral.atom1.id_in_mol
            idx2 = dihedral.atom2.id_in_mol
            idx3 = dihedral.atom3.id_in_mol
            idx4 = dihedral.atom4.id_in_mol
            mol.add_dihedral(mol._atoms[idx1], mol._atoms[idx2], mol._atoms[idx3], mol._atoms[idx4])
        for improper in self._impropers:
            idx1 = improper.atom1.id_in_mol
            idx2 = improper.atom2.id_in_mol
            idx3 = improper.atom3.id_in_mol
            idx4 = improper.atom4.id_in_mol
            mol.add_improper(mol._atoms[idx1], mol._atoms[idx2], mol._atoms[idx3], mol._atoms[idx4])

        # add_atom and add_bond will invalidate rdmol
        if self._is_rdmol_valid:
            from rdkit import Chem
            mol._rdmol = Chem.Mol(self._rdmol)
            mol._is_rdmol_valid = True

        for i, atom in enumerate(self._atoms):
            vsite = atom.virtual_site
            if vsite is not None:
                new_atom = mol.atoms[i]
                new_parents = [mol.atoms[p.id_in_mol] for p in vsite.parents]
                new_atom.virtual_site = VirtualSite.create(vsite.__class__.__name__, new_parents, vsite.parameters)

        for residue in self._added_residues:
            atoms = [mol.atoms[atom.id_in_mol] for atom in residue.atoms]
            mol.add_residue(residue.name, atoms, refresh_residues=False)
        mol.refresh_residues()

        return mol

    @staticmethod
    def from_smiles(smiles):
        '''
        Initialize a molecule from SMILES string.

        RDKit is used for parsing SMILES. The Hydrogen atoms will be created.
        The positions of all atoms will also be automatically generated.
        The SMILES string can contain the name of the molecule at the end, e.g. 'CCCC butane'.

        Parameters
        ----------
        smiles : str

        Returns
        -------
        molecule : Molecule
        '''
        words = smiles.strip().split()
        if not words:
            raise Exception('Invalid SMILES string')

        smiles = words[0]
        if len(words) > 1:
            name = words[1]
        else:
            name = None

        rdmol = create_mol_from_smiles(smiles)
        mol = Molecule.from_rdmol(rdmol, name)

        return mol

    @staticmethod
    def from_rdmol(rdmol, name=None):
        '''
        Initialize a molecule from a RDKit Mol object.
        If the RDKit Mol has conformers, the position of the first conformer will be assigned to the atoms

        Parameters
        ----------
        rdmol : rdkit.Chem.Mol
        name : str
            The name of the molecule. If not provided, the formula will be used as the name.

        Returns
        -------
        molecule : Molecule
        '''
        try:
            from rdkit import Chem
            from rdkit.Chem.rdMolDescriptors import CalcMolFormula
        except ImportError:
            raise ImportError('RDKit not found')

        rdmol = Chem.Mol(rdmol)
        # don't set aromaticity, kekulized bonds are easier to manipulate
        Chem.SanitizeMol(rdmol, Chem.SANITIZE_ALL ^ Chem.SANITIZE_SETAROMATICITY)
        mol = Molecule()
        for i, a in enumerate(rdmol.GetAtoms()):
            atom = Atom()
            element = Element(a.GetAtomicNum())
            atom.name = element.symbol + str(i + 1)
            atom.symbol = element.symbol
            atom.mass = element.mass
            atom.formal_charge = a.GetFormalCharge()
            mol.add_atom(atom)
        if rdmol.GetNumConformers() > 0:
            for atom, pos in zip(mol.atoms, rdmol.GetConformer().GetPositions()):
                atom.position = pos / 10  # convert A to nm
        for b in rdmol.GetBonds():
            atom1 = mol.atoms[b.GetBeginAtomIdx()]
            atom2 = mol.atoms[b.GetEndAtomIdx()]
            d_bond_order = {
                Chem.rdchem.BondType.UNSPECIFIED: Bond.Order.UNSPECIFIED,
                Chem.rdchem.BondType.SINGLE     : Bond.Order.SINGLE,
                Chem.rdchem.BondType.DOUBLE     : Bond.Order.DOUBLE,
                Chem.rdchem.BondType.TRIPLE     : Bond.Order.TRIPLE,
            }
            try:
                order = d_bond_order[b.GetBondType()]
            except KeyError:
                logger.warning('Only single/double/triple/aromatic bond supported. Will discard bond order')
                order = Bond.Order.UNSPECIFIED
            mol.add_bond(atom1, atom2, order)
        mol.generate_angle_dihedral_improper()

        # set aromaticity so that SMARTS matching and is_aromatic works correctly
        Chem.SetAromaticity(rdmol)
        mol._rdmol = rdmol
        mol._is_rdmol_valid = True

        if name is not None:
            mol.name = name
        else:
            mol.name = CalcMolFormula(rdmol)

        return mol

    @property
    def rdmol(self):
        '''
        The `rdkit.Chem.Mol` object associated with this molecule.

        It is required by ZftTyper typing engine, which performs SMARTS matching on the molecule.
        The `rdmol` attribute will be assigned if the molecule is initialized from SMILES or RDKit Molecule.
        If it is not available, a RDKit molecule will be constructed from atoms and bonds.
        The positions will not be preserved.

        Returns
        -------
        rdmol : rdkit.Chem.Mol
        '''
        try:
            from rdkit import Chem
        except ImportError:
            raise ImportError('RDKit not found')

        if self._is_rdmol_valid:
            return self._rdmol

        if any(b.order == Bond.Order.UNSPECIFIED for b in self.bonds):
            logger.warning(f'Not all bond orders are specified in {self}')

        rwmol = Chem.RWMol()
        for atom in self.atoms:
            rdatom = Chem.Atom(atom.symbol)
            rdatom.SetFormalCharge(atom.formal_charge)
            rdatom.SetNoImplicit(True)  # disable implicit Hs. Otherwise cannot handle radicals
            rwmol.AddAtom(rdatom)
        for bond in self.bonds:
            d_bond_order = {
                Bond.Order.UNSPECIFIED: Chem.rdchem.BondType.UNSPECIFIED,
                Bond.Order.SINGLE     : Chem.rdchem.BondType.SINGLE,
                Bond.Order.DOUBLE     : Chem.rdchem.BondType.DOUBLE,
                Bond.Order.TRIPLE     : Chem.rdchem.BondType.TRIPLE,
            }
            rwmol.AddBond(bond.atom1.id_in_mol, bond.atom2.id_in_mol, d_bond_order[bond.order])
        # set aromaticity so that SMARTS matching and is_aromatic works correctly
        Chem.SanitizeMol(rwmol)
        self._rdmol = rwmol.GetMol()
        self._is_rdmol_valid = True

        return self._rdmol

    def generate_conformers(self, n_conformer=1):
        '''
        Generate several conformers with RDKit.

        The positions will be generated from only elements and bonds. The chiral center will not be respected.

        Parameters
        ----------
        n_conformer : int
            How many conformers to generate

        Returns
        -------
        positions_list : list of array_like
        '''
        try:
            from rdkit.Chem import AllChem as Chem
        except ImportError:
            raise ImportError('RDKit not found')

        positions_list = []
        rdmol = Chem.Mol(self.rdmol)
        useRandomCoords = False
        while len(positions_list) < n_conformer:
            Chem.EmbedMultipleConfs(rdmol, numConfs=n_conformer - len(positions_list), clearConfs=True,
                                    useRandomCoords=useRandomCoords)
            for i in range(rdmol.GetNumConformers()):
                positions_list.append(rdmol.GetConformer(i).GetPositions() / 10)  # A -> nm
            useRandomCoords = True

        return positions_list

    @property
    def topology(self):
        '''
        The topology this molecule belongs to

        Returns
        -------
        topology : Topology

        '''
        return self._topology

    def add_atom(self, atom, residue=None, index=None, update_topology=True):
        '''
        Add an atom to this molecule.

        The id_in_mol attribute of all atoms will be updated after insertion.

        TODO Make residue assignment more robust

        Parameters
        ----------
        atom : Atom
        residue : Residue, Optional
            Add the atom to this residue.
            Make sure the residue belongs to this molecule. For performance concern, this is not checked.
            If set to None, the atom will be added to the default residue.
        index : int, Optional
            If None, the new atom will be the last atom. Otherwise, it will be inserted in front of index-th atom.
        update_topology : bool
            If True, the topology this molecule belongs to will update its atom list and assign id for all atoms and residues.
            Otherwise, you have to re-init the topology manually so that the topological information is correct.
        '''
        atom._molecule = self
        if index is None:
            self._atoms.append(atom)
            atom.id_in_mol = len(self._atoms) - 1
        else:
            self._atoms.insert(index, atom)
            for i in range(index, len(self._atoms)):
                self._atoms[i].id_in_mol = i
        self._is_rdmol_valid = False

        residue = residue or self._default_residue
        residue._add_atom(atom)
        # re-index residues because the residue starts to count
        if residue.n_atom == 1:
            self.refresh_residues(update_topology)

        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

    def remove_atom(self, atom, update_topology=True):
        '''
        Remove an atom and all the bonds connected to the atom from this molecule.

        The atom will also be removed from its residue.
        The id_in_mol attribute of all atoms will be updated after removal.
        The angle, dihedral and improper involving this atom are untouched.
        Therefore, you may call `generate_angle_dihedral_improper` to refresh the connectivity.

        Parameters
        ----------
        atom : Atom
        update_topology : bool
            If update_topology is True, the topology this molecule belongs to will update its atom list and assign id for all atoms and residues.
            Otherwise, you have to re-init the topology manually so that the topological information is correct.
        '''
        self.remove_atoms([atom], update_topology=update_topology)

    def remove_atoms(self, atoms, update_topology=True):
        '''
        Remove multiple atoms and all the bonds connected to these atoms from this molecule.

        The atom will also be removed from its residue.
        The id_in_mol attribute of all atoms will be updated after removal.
        The angle, dihedral and improper involving this atom are untouched.
        Therefore, you may call `generate_angle_dihedral_improper` to refresh the connectivity.

        Parameters
        ----------
        atom : Atom
        update_topology : bool
            If update_topology is True, the topology this molecule belongs to will update its atom list and assign id for all atoms and residues.
            Otherwise, you have to re-init the topology manually so that the topological information is correct.
        '''
        _refresh_residues = False
        for atom in atoms:
            for bond in atom._bonds[:]:
                self.remove_connectivity(bond)
            # self._atoms.remove(atom)
            atom._molecule = None

            residue = atom.residue
            residue._remove_atom(atom)
            if not _refresh_residues and residue.n_atom == 0:
                _refresh_residues = True

        # it's faster to reconstruct atom list than iteratively calling _atoms.remove()
        self._atoms = list(sorted(set(self._atoms) - set(atoms), key=lambda x: x.id_in_mol))

        for i, at in enumerate(self._atoms):
            at.id_in_mol = i
        self._is_rdmol_valid = False

        # re-index residues because some residues become empty
        if _refresh_residues:
            self.refresh_residues(update_topology)

        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

    def remove_non_polar_hydrogens(self, update_topology=True):
        '''
        Remove single-coordinated hydrogen atoms bonded to C and Si atoms

        Parameters
        ----------
        update_topology : bool
            If update_topology is True, the topology this molecule belongs to will update its atom list and assign id for all atoms and residues.
            Otherwise, you have to re-init the topology manually so that the topological information is correct.

        Returns
        -------
        ids_removed : list of int
            The list of `id_in_mol` of atoms removed
        '''
        hydrogens = []
        for atom in self.atoms[:]:
            if atom.symbol != 'H' or len(atom.bonds) != 1:
                continue
            neigh = atom.bond_partners[0]
            if neigh.symbol not in ('C', 'Si'):
                continue
            neigh.mass += atom.mass
            neigh.charge += atom.charge
            hydrogens.append(atom)

        ids_hydrogens = [atom.id_in_mol for atom in hydrogens]
        set_hydrogens = set(hydrogens)

        # cannot modify _atoms and _bonds directly because there are internal data structures to maintain
        # have to call remove_atoms()
        # remove_atoms() will also remove relevant bonds
        self.remove_atoms(hydrogens, update_topology=False)

        # it's faster to reconstruct angle list than iteratively calling _angles.remove()
        self._angles = [conn for conn in self.angles if not set(conn.atoms).intersection(set_hydrogens)]
        self._dihedrals = [conn for conn in self.dihedrals if not set(conn.atoms).intersection(set_hydrogens)]
        self._impropers = [conn for conn in self.impropers if not set(conn.atoms).intersection(set_hydrogens)]

        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

        return ids_hydrogens

    def add_residue(self, name, atoms, refresh_residues=True):
        '''
        Put a group of atoms into a new residue. These atoms will be removed from their old residues.

        Make sure that these atoms belong to this molecule.
        For performance issue, this is not checked.

        Parameters
        ----------
        name : str
        atoms : list of Atom
        refresh_residues: bool
            If True, the residue list of this molecule will be updated. The `id` and `id_in_mol` for all residues will be assigned.
            This operation can be slow. If you have a lot of residues to add, it is more efficient to set it to False,
            and call `refresh_residues` manually after all residues are added.

        Returns
        -------
        residue : Residue
        '''
        residue = Residue(name)
        for atom in atoms:
            atom.residue._remove_atom(atom)
            residue._add_atom(atom)
        self._added_residues.append(residue)
        if refresh_residues:
            self.refresh_residues()

        return residue

    def remove_residue(self, residue, refresh_residues=True):
        '''
        Remove a residue from this molecule, and put the relevant atoms into the default residue

        Make sure that this residue belongs to this molecule.
        For performance issue, this is not checked.

        Parameters
        ----------
        residue : Residue
        refresh_residues: bool
            If True, the residue list of this molecule will be updated. The `id` and `id_in_mol` for all residues will be assigned.
            This operation can be slow. If you have a lot of residues to remove, it is more efficient to set it to False,
            and call `refresh_residues` manually after all residues are removed.
        '''
        for atom in residue.atoms[:]:
            self._default_residue._add_atom(atom)
        self._added_residues.remove(residue)
        if refresh_residues:
            self.refresh_residues()

    def refresh_residues(self, update_topology=True):
        '''
        Remove empty residues, update `id_in_mol` attributes of each residue in this molecule

        Parameters
        ----------
        update_topology : bool
            If True, the topology this molecule belongs to will assign `id` for all residues
        '''
        self._added_residues = [res for res in self._added_residues if res.n_atom != 0]
        for i, residue in enumerate(self.residues):
            residue.id_in_mol = i
        if self._topology is not None and update_topology:
            for i, residue in enumerate(self._topology.residues):
                residue.id = i

    def add_bond(self, atom1, atom2, order=Bond.Order.UNSPECIFIED, check_existence=False):
        '''
        Add a bond between two atoms.

        Make sure that both these two atoms belong to this molecule.
        For performance issue, this is not checked.

        Parameters
        ----------
        atom1 : Atom
        atom2 : Atom
        check_existence : bool
            If set to True and there is already bond between these two atoms, then do nothing and return None

        Returns
        -------
        bond : [Bond, None]
        '''
        bond = Bond(atom1, atom2, order)
        if check_existence and any(b.equals(bond) for b in self._bonds):
            return None

        self._bonds.append(bond)
        atom1._bonds.append(bond)
        atom2._bonds.append(bond)
        self._is_rdmol_valid = False

        return bond

    def add_angle(self, atom1, atom2, atom3, check_existence=False):
        '''
        Add a angle between three atoms.

        The second atom is the central atom.
        Make sure that both these three atoms belong to this molecule.
        For performance issue, this is not checked.

        Parameters
        ----------
        atom1 : Atom
        atom2 : Atom
        atom3 : Atom
        check_existence : bool
            If set to True and there is already angle between these three atoms, then do nothing and return None
        Returns
        -------
        angle : [Angle, None]
        '''
        angle = Angle(atom1, atom2, atom3)
        if check_existence and any(a.equals(angle) for a in self._angles):
            return None

        self._angles.append(angle)
        return angle

    def add_dihedral(self, atom1, atom2, atom3, atom4, check_existence=False):
        '''
        Add a dihedral between four atoms.

        Make sure that both these four atoms belong to this molecule.
        For performance issue, this is not checked.

        Parameters
        ----------
        atom1 : Atom
        atom2 : Atom
        atom3 : Atom
        atom4 : Atom
        check_existence : bool
            If set to True and there is already dihedral between these three atoms, then do nothing and return None
        Returns
        -------
        dihedral : [Dihedral, None]
        '''
        dihedral = Dihedral(atom1, atom2, atom3, atom4)
        if check_existence and any(d.equals(dihedral) for d in self._dihedrals):
            return None

        self._dihedrals.append(dihedral)
        return dihedral

    def add_improper(self, atom1, atom2, atom3, atom4, check_existence=False):
        '''
        Add a improper between four atoms.

        The fist atom is the central atom.
        Make sure that both these four atoms belong to this molecule.
        For performance issue, this is not checked.

        Parameters
        ----------
        atom1 : Atom
        atom2 : Atom
        atom3 : Atom
        atom4 : Atom
        check_existence : bool
            If set to True and there is already improper between these three atoms, then do nothing and return None
        Returns
        -------
        dihedral : [Improper, None]
        '''
        improper = Improper(atom1, atom2, atom3, atom4)
        if check_existence and any(i.equals(improper) for i in self._impropers):
            return None

        self._impropers.append(improper)
        return improper

    def remove_connectivity(self, connectivity):
        '''
        Remove a connectivity (bond, angle, diheral or improper) from this molecule.

        Make sure that this connectivity belongs to this molecule.
        For performance issue, this is not checked.

        Note that when a bond get removed, the relevant angles, dihedrals and impropers are still there.
        You may call `generate_angle_dihedral_improper` to refresh connectivity.

        # TODO This operation is slow. A batch version is required for better performance

        Parameters
        ----------
        connectivity : [Bond, Angle, Dihedral, Improper]
        '''
        if type(connectivity) is Bond:
            bond = connectivity
            self._bonds.remove(bond)
            bond.atom1._bonds.remove(bond)
            bond.atom2._bonds.remove(bond)
            self._is_rdmol_valid = False
        elif type(connectivity) is Angle:
            self._angles.remove(connectivity)
        elif type(connectivity) is Dihedral:
            self._dihedrals.remove(connectivity)
        elif type(connectivity) is Improper:
            self._impropers.remove(connectivity)
        else:
            raise Exception('Invalid connectivity')

    def is_similar_to(self, other):
        '''
        Check if this molecule is similar to another molecule.

        It requires two molecules contains the same number of atoms.
        The correspond atoms should have same atom symbol, type and charge.
        The bonds should also be the same.
        But it doesn't consider angles, dihedrals and impropers.

        Parameters
        ----------
        other : Molecule

        Returns
        -------
        is : bool
        '''
        other: Molecule
        if self.n_atom != other.n_atom:
            return False
        if self.n_bond != other.n_bond:
            return False
        for i in range(self.n_atom):
            atom1 = self._atoms[i]
            atom2 = other._atoms[i]
            if atom1.symbol != atom2.symbol or atom1.type != atom2.type or atom1.charge != atom2.charge:
                return False
            if len(atom1._bonds) != len(atom2._bonds):
                return False
            if set(p.id_in_mol for p in atom1.bond_partners) != \
                    set(p.id_in_mol for p in atom2.bond_partners):
                return False
        return True

    def get_adjacency_matrix(self):
        matrix = np.zeros([self.n_atom, self.n_atom], dtype=bool)
        for bond in self.bonds:
            a1, a2 = bond.atom1.id_in_mol, bond.atom2.id_in_mol
            matrix[a1][a2] = True
            matrix[a2][a1] = True

        # no bond between virtual sites and parents
        for vsite, parent in self.get_virtual_site_pairs():
            a1, a2 = vsite.id_in_mol, parent.id_in_mol
            matrix[a1][a2] = True
            matrix[a2][a1] = True

        return matrix

    def get_distance_matrix(self, max_bond=None):
        connections = [set() for _ in range(self.n_atom)]
        for bond in self._bonds:
            a1, a2 = bond.atom1.id_in_mol, bond.atom2.id_in_mol
            connections[a1].add(a2)
            connections[a2].add(a1)
        mat = np.zeros((self.n_atom, self.n_atom), dtype=int)

        def fill_matrix(level, center, neighbors, flags):
            if max_bond and level > max_bond:
                return

            if all(flags):
                return

            for j in neighbors:
                if not flags[j]:
                    mat[center][j] = level
                    mat[j][center] = level
                    flags[j] = True

            neighbors_deeper = set()
            for j in neighbors:
                neighbors_deeper.update(connections[j])
            fill_matrix(level + 1, center, neighbors_deeper, flags)

        for i in range(self.n_atom):
            fill_matrix(1, i, connections[i], [j == i for j in range(self.n_atom)])

        return mat

    @property
    def n_atom(self):
        '''
        Number of atoms belong to this molecule

        Returns
        -------
        n : int
        '''
        return len(self._atoms)

    @property
    def n_bond(self):
        '''
        Number of bonds belong to this molecule

        Returns
        -------
        n : int
        '''
        return len(self._bonds)

    @property
    def n_angle(self):
        '''
        Number of angles belong to this molecule

        Returns
        -------
        n : int
        '''
        return len(self._angles)

    @property
    def n_dihedral(self):
        '''
        Number of dihedrals belong to this molecule

        Returns
        -------
        n : int
        '''
        return len(self._dihedrals)

    @property
    def n_improper(self):
        '''
        Number of impropers belong to this molecule

        Returns
        -------
        n : int
        '''
        return len(self._impropers)

    @property
    def n_residue(self):
        return len(self.residues)

    @property
    def atoms(self):
        '''
        List of atoms belong to this molecule

        Returns
        -------
        atoms: list of Atom
        '''
        return self._atoms

    @property
    def bonds(self):
        '''
        List of bonds belong to this molecule

        Returns
        -------
        bonds : list of Bond
        '''
        return self._bonds

    @property
    def angles(self):
        '''
        List of angles belong to this molecule

        Returns
        -------
        angles : list of Angle
        '''
        return self._angles

    @property
    def dihedrals(self):
        '''
        List of dihedrals belong to this molecule

        Returns
        -------
        dihedrals : list of Dihedral
        '''
        return self._dihedrals

    @property
    def impropers(self):
        '''
        List of impropers belong to this molecule

        Returns
        -------
        impropers : list of Improper
        '''
        return self._impropers

    @property
    def residues(self):
        '''
        All the residues in this molecule

        Returns
        -------
        residues : list of Residue
        '''
        if len(self._default_residue.atoms) == 0:
            return self._added_residues
        else:
            return [self._default_residue] + self._added_residues

    @property
    def has_position(self):
        '''
        Whether or not all the atoms in the molecule have positions

        Returns
        -------
        has : bool
        '''
        return all(atom.has_position for atom in self.atoms)

    @property
    def positions(self):
        '''
        Positions of all the atoms in this molecule

        Returns
        -------
        positions : array_like
        '''
        return np.array([atom.position for atom in self.atoms])

    def set_positions(self, positions):
        '''
        Set the positions of all atoms in this molecule

        Parameters
        ----------
        positions : array_like
        '''
        if self.n_atom != len(positions):
            raise Exception('Length of positions should equal to the number of atoms')
        for i, atom in enumerate(self.atoms):
            atom.position = positions[i]

    def get_drude_pairs(self):
        '''
        Retrieve all the Drude dipole pairs belong to this molecule

        Returns
        -------
        pairs :  list of tuple of Atom
            [(parent, drude)]
        '''
        pairs = []
        for bond in self._bonds:
            if bond.atom1.is_drude:
                pairs.append((bond.atom2, bond.atom1))
            elif bond.atom2.is_drude:
                pairs.append((bond.atom1, bond.atom2))
        return pairs

    def get_virtual_site_pairs(self):
        '''
        Retrieve all the virtual site pairs belong to this molecule

        Returns
        -------
        pairs :  list of tuple of Atom
            [(parent, atom_virtual_site)]
        '''
        pairs = []
        for atom in self._atoms:
            if atom.virtual_site is not None:
                pairs.append((atom.virtual_site.parents[0], atom))
        return pairs

    def get_12_13_14_pairs(self):
        '''
        Retrieve all the 1-2, 1-3 and 1-4 pairs based on the bond information.

        The pairs only concerns real atoms. Drude particles  will be ignored.

        Returns
        -------
        pairs12 : list of tuple of Atom
        pairs13 : list of tuple of Atom
        pairs14 : list of tuple of Atom
        '''
        pair_12_set = set()
        pair_13_set = set()
        pair_14_set = set()

        for atom in [a for a in self._atoms if not a.is_drude]:
            partners = [p for p in atom.bond_partners if not p.is_drude]
            for a1, a3 in itertools.combinations(partners, 2):
                pair = tuple(sorted([a1, a3]))
                pair_13_set.add(pair)

        for bond in filter(lambda x: not x.is_drude, self.bonds):
            a2, a3 = bond.atom1, bond.atom2
            pair = tuple(sorted([a2, a3]))
            pair_12_set.add(pair)
            for a1 in [p for p in a2.bond_partners if not p.is_drude]:
                for a4 in [p for p in a3.bond_partners if not p.is_drude]:
                    if a1 != a3 and a2 != a4 and a1 != a4:
                        pair = tuple(sorted([a1, a4]))
                        pair_14_set.add(pair)

        pair_12_list = list(sorted(pair_12_set))
        pair_13_list = list(sorted(pair_13_set - pair_12_set))
        pair_14_list = list(sorted(pair_14_set - pair_13_set.union(pair_12_set)))
        return pair_12_list, pair_13_list, pair_14_list

    def generate_angle_dihedral_improper(self, dihedral=True, improper=True):
        '''
        Generate angle, dihedral and improper from bonds
        The existing angles, dihedrals and impropers will be removed first
        The atoms and bonds concerning Drude particles will be ignored

        Parameters
        ----------
        dihedral: bool
            Whether or not generate dihedrals based on bonds
        improper: bool
            Whether or not generate impropers based on bonds
        '''
        self._angles = []
        self._dihedrals = []
        self._impropers = []

        for atom in [a for a in self._atoms if not a.is_drude]:
            partners = [p for p in atom.bond_partners if not p.is_drude]
            for p1, p2 in itertools.combinations(partners, 2):
                self.add_angle(p1, atom, p2)
            if improper and len(partners) == 3:
                self.add_improper(atom, *sorted(partners))

        if dihedral:
            for bond in filter(lambda x: not x.is_drude, self._bonds):
                atom2 = bond.atom1
                atom3 = bond.atom2
                partners2 = [p for p in atom2.bond_partners if not p.is_drude]
                partners3 = [p for p in atom3.bond_partners if not p.is_drude]
                for atom1, atom4 in itertools.product(partners2, partners3):
                    if atom1 != atom3 and atom2 != atom4 and atom1 != atom4:
                        self.add_dihedral(atom1, atom2, atom3, atom4)

    def guess_connectivity_from_ff(self, ff, bond_limit=0.25, bond_tolerance=0.025, angle_tolerance=None,
                                   pbc='', cell=None):
        '''
        Guess bonds, angles, dihedrals and impropers from force field.

        It requires that atoms types are defined and positions are available.
        The distance between nearby atoms will be calculated.
        If it's smaller than bond_length_limit, then it will be compared with the equilibrium length in FF.
        The bond will be added if a BondTerm is found in FF and the deviation is smaller than bond_tolerance.
        Then angles will be constructed from bonds. If angle_tolerance is None, all angles will be added.
        If angle_tolerance is set (as degree), then AngleTerm must be provided for these angles.
        The angle will be added only if the deviation between angle and equilibrium value in FF is smaller than angle_tolerance.
        Dihedrals and impropers will be constructed form bonds and be added if relevant terms are presented in FF.

        PBC is supported for determining bonds across the periodic cell
        This is useful for simulating infinite structures
        pbc can be '', 'x', 'y', 'xy', 'xz', 'xyz', which means check bonds cross specific boundaries
        cell should also be provided if pbc is not ''

        TODO Add support for triclinic cell

        Parameters
        ----------
        ff : ForceField
        bond_limit : float
        bond_tolerance : float
        angle_tolerance : float
        pbc : str
        cell : UnitCell
        '''
        if not self.has_position:
            raise Exception('Positions are required for guessing connectivity')

        if any(atom.is_drude for atom in self._atoms):
            raise Exception('Drude particles should be removed before guess connectivity')

        if pbc != '':
            if cell is None or cell.volume == 0:
                raise Exception('PBC required but valid cell not provided')
            elif not cell.is_rectangular:
                raise Exception('Triclinic cell haven\'t been implemented')
            else:
                box = cell.size

        self._bonds = []
        self._angles = []
        self._dihedrals = []
        self._impropers = []

        for i in range(self.n_atom):
            atom1 = self.atoms[i]
            try:
                at1 = ff.atom_types[atom1.type].eqt_bond
            except:
                raise Exception(f'AtomType {atom1.type} not found in FF')

            for j in range(i, self.n_atom):
                atom2 = self.atoms[j]
                try:
                    at2 = ff.atom_types[atom2.type].eqt_bond
                except:
                    raise Exception(f'AtomType {atom2.type} not found in FF')

                delta = atom2.position - atom1.position
                if pbc != '':
                    if any((np.abs(delta) > bond_limit) & (np.abs(delta) < box - bond_limit)):
                        continue
                    if 'x' in pbc:
                        delta[0] -= math.ceil(delta[0] / box[0] - 0.5) * box[0]
                    if 'y' in pbc:
                        delta[1] -= math.ceil(delta[1] / box[1] - 0.5) * box[1]
                    if 'z' in pbc:
                        delta[2] -= math.ceil(delta[2] / box[2] - 0.5) * box[2]
                if any(np.abs(delta) > bond_limit):
                    continue

                bterm = BondTerm(at1, at2, 0)
                if bterm.name not in ff.bond_terms.keys():
                    continue

                bterm: BondTerm = ff.bond_terms[bterm.name]
                if any(delta - bterm.length > bond_tolerance):
                    continue

                if abs(np.sqrt(delta.dot(delta)) - bterm.length) <= bond_tolerance:
                    self.add_bond(atom1, atom2)

        # generate angles etc..., and then remove them if requirements are not satisfied
        self.generate_angle_dihedral_improper()

        angles_removed = []
        dihedrals_removed = []
        impropers_removed = []
        if angle_tolerance is not None:
            for angle in self._angles[:]:
                at1 = ff.atom_types[angle.atom1.type].eqt_ang_s
                at2 = ff.atom_types[angle.atom2.type].eqt_ang_c
                at3 = ff.atom_types[angle.atom3.type].eqt_ang_s
                aterm = AngleTerm(at1, at2, at3, 0)
                if aterm.name not in ff.angle_terms.keys():
                    raise Exception(
                        f'{str(angle)} constructed but {str(aterm)} not found in FF')

                aterm: AngleTerm = ff.angle_terms[aterm.name]
                delta21 = angle.atom1.position - angle.atom2.position
                delta23 = angle.atom3.position - angle.atom2.position
                if 'x' in pbc:
                    delta21[0] -= math.ceil(delta21[0] / box[0] - 0.5) * box[0]
                    delta23[0] -= math.ceil(delta23[0] / box[0] - 0.5) * box[0]
                if 'y' in pbc:
                    delta21[1] -= math.ceil(delta21[1] / box[1] - 0.5) * box[1]
                    delta23[1] -= math.ceil(delta23[1] / box[1] - 0.5) * box[1]
                if 'z' in pbc:
                    delta21[2] -= math.ceil(delta21[2] / box[2] - 0.5) * box[2]
                    delta23[2] -= math.ceil(delta23[2] / box[2] - 0.5) * box[2]
                cos = delta21.dot(delta23) / math.sqrt(delta21.dot(delta21) * delta23.dot(delta23))
                theta = np.arccos(np.clip(cos, -1, 1))
                if abs(theta - aterm.theta) > angle_tolerance * DEG2RAD:
                    self.remove_connectivity(angle)
                    angles_removed.append(angle)

        for dihedral in self._dihedrals[:]:
            # consider wildcards in force field
            ats_list = ff.get_eqt_for_dihedral(dihedral)
            for ats in ats_list:
                dterm = DihedralTerm(*ats)
                if dterm.name in ff.dihedral_terms.keys():
                    break
            else:
                self.remove_connectivity(dihedral)
                dihedrals_removed.append(dihedral)

        for improper in self._impropers[:]:
            # consider wildcards in force field
            ats_list = ff.get_eqt_for_improper(improper)
            for ats in ats_list:
                iterm = ImproperTerm(*ats)
                if iterm.name in ff.improper_terms.keys():
                    break
            else:
                self.remove_connectivity(improper)
                impropers_removed.append(improper)

        if angles_removed != []:
            msg = '%i angles not added because value far from equilibrium: ' \
                  % len(angles_removed) \
                  + ' '.join([i.name for i in angles_removed[:10]])
            if len(angles_removed) > 10:
                msg += ' and more ...'
            logger.warning(msg)

        if dihedrals_removed != []:
            msg = '%i dihedrals not added because parameters not found in FF: ' \
                  % len(dihedrals_removed) \
                  + ' '.join([i.name for i in dihedrals_removed[:10]])
            if len(dihedrals_removed) > 10:
                msg += ' and more ...'
            logger.warning(msg)

        if impropers_removed != []:
            msg = '%i impropers not added because parameters not found in FF: ' \
                  % len(impropers_removed) \
                  + ' '.join([i.name for i in impropers_removed[:10]])
            if len(impropers_removed) > 10:
                msg += ' and more ...'
            logger.warning(msg)

    def generate_drude_particles(self, ff, type_drude='DP_', seed=1, update_topology=True):
        '''
        Generate Drude particles from DrudeTerms in force field.

        The atom types should have been defined already.
        Drude particle will not be generated if DrudeTerm for its atom type can not be found in the FF.
        Note that The existing Drude particles will be removed before generating.
        The mass defined in the DrudeTerm will be transferred from parent atom to the Drude particle.
        The Drude charge will be calculated from the DrudeTerm and transferred from parent atom to the Drude particle.
        Bonds between parent-Drude will be generated and added to the topology.
        If AtomType and VdwTerm for generated Drude particles are not found in FF, these terms will be created and added to the FF.

        Parameters
        ----------
        ff : ForceField
        type_drude : str
        seed : int
        update_topology : bool
        '''
        if len(ff.polarizable_terms) == 0:
            raise Exception('Polarizable terms not found in force field')

        np.random.seed(seed)

        self.remove_drude_particles(update_topology=False)

        _atype_not_found = set()
        drude_pairs = {}
        for parent in self._atoms:
            atype = ff.atom_types.get(parent.type)
            if atype is None:
                _atype_not_found.add(parent.type)
                continue

            pterm = ff.polarizable_terms.get(atype.eqt_polar)
            if pterm is None:
                continue
            if type(pterm) is not DrudeTerm:
                raise Exception('Polarizable terms other than DrudeTerm haven\'t been implemented')
            drude = Atom()
            drude.is_drude = True
            drude.type = type_drude
            # add Drude particles after all been generated so the name of them are in sequence
            drude.name = 'DP' + str(parent.id_in_mol + 1)
            drude.symbol = 'DP'
            drude.mass = pterm.mass
            parent.mass -= drude.mass
            n_H = len([atom for atom in parent.bond_partners if atom.symbol == 'H'])
            alpha = pterm.alpha + n_H * pterm.merge_alpha_H
            drude.charge = - pterm.get_charge(alpha)
            parent.charge += pterm.get_charge(alpha)
            # update alpha and thole for Drude parent particle
            parent.alpha = alpha
            parent.thole = pterm.thole

            if parent.has_position:
                # make sure Drude and parent atom do not overlap. max deviation 0.005 nm
                drude.position = parent.position + (np.random.random(3) - 0.5) / 100

            drude_pairs[parent] = drude

        if _atype_not_found != set():
            logger.error('%i atom types not found in FF: %s' % (
                len(_atype_not_found), ' '.join(_atype_not_found)))
            raise Exception(f'Generating Drude particles {str(self)} failed')

        for parent, drude in drude_pairs.items():
            self.add_atom(drude, index=self._atoms.index(parent) + 1, update_topology=False)
            self.add_bond(parent, drude)

        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

        dtype = ff.atom_types.get(type_drude)
        if dtype is None:
            dtype = AtomType(type_drude)
            ff.add_term(dtype)
            logger.warning(f'AtomType for Drude particle not found in FF. '
                           f'{str(dtype)} is added to the FF')
        vdw = LJ126Term(dtype.eqt_vdw, dtype.eqt_vdw, 0.0, 0.0)
        if ff.vdw_terms.get(vdw.name) is None:
            ff.add_term(vdw)
            logger.warning(f'VdwTerm for Drude particle not found in FF. '
                           f'{str(vdw)} with zero interactions is added to the FF')

        for atom in self._atoms:
            if not atom.is_drude and atom.symbol != 'H' and atom not in drude_pairs:
                logger.warning(f'Not all heavy atoms in {str(self)} carry Drude particles')
                break

    def remove_drude_particles(self, update_topology=True):
        '''
        Remove all Drude particles and bonds belong to Drude particles

        The charges and masses carried by Drude particles will be transferred back to parent atoms

        Parameters
        ----------
        update_topology : bool
        '''
        for parent, drude in self.get_drude_pairs():
            parent.mass += drude.mass
            parent.charge += drude.charge
            self.remove_connectivity(drude._bonds[0])
            self.remove_atom(drude, update_topology=False)
        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

    def generate_virtual_sites(self, ff, update_topology=True):
        '''
        Generate virtual sites from VirtualSiteTerms in force field.

        The atom types should have been defined already.
        Note that The existing virtual sites will be removed before generating.
        The charge won't be assigned by this method.
        Therefore `assign_charge_from_ff` should be called to assign the charges on virtual sites.

        Currently, only TIP4PSiteTerm has been implemented.

        TODO Support other virtual site terms

        Parameters
        ----------
        ff : ForceField
        update_topology : bool
        '''
        if len(ff.virtual_site_terms) == 0:
            raise Exception('Virtual site terms not found in force field')

        self.remove_virtual_sites(update_topology=False)

        for term in ff.virtual_site_terms.values():
            if type(term) is not TIP4PSiteTerm:
                raise Exception('Virtual sites terms other than TIP4PSiteTerm haven\'t been implemented')

        for term in ff.virtual_site_terms.values():
            for angle in self._angles:
                if angle.atom2.type != term.type_O or angle.atom1.type != term.type_H or angle.atom3.type != term.type_H:
                    continue
                atom_vsite = Atom('VS' + str(angle.atom2.id_in_mol + 1))
                atom_vsite.symbol = 'VS'
                atom_vsite.type = term.type
                atom_vsite.virtual_site = TIP4PSite([angle.atom2, angle.atom1, angle.atom3], [term.d])
                atom_vsite.position = atom_vsite.virtual_site.calc_position()
                self.add_atom(atom_vsite, update_topology=False)

        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

    def remove_virtual_sites(self, update_topology=True):
        '''
        Remove all virtual sites.

        Parameters
        ----------
        update_topology : bool
        '''
        for atom in self._atoms:
            if atom.virtual_site is not None:
                self.remove_atom(atom, update_topology=False)
        if self._topology is not None and update_topology:
            self._topology.update_molecules(self._topology.molecules, deepcopy=False)

    def get_sub_molecule(self, indexes, deepcopy=True):
        '''
        Extract a substructure from this molecule by indexes of atoms.

        The substructure will not contain any bond, angle, dihedral and improper between atoms in substructure and remaining parts.
        Residue information will be reconstructed.

        TODO Fix performance issue

        Parameters
        ----------
        indexes : list of int
            The atoms in the substructure will be in the same order as in indexes
        deepcopy : bool
            If set to False, then the atoms and connections in the substructure will be the identical object as the atoms and connections in this molecule.
            The data structure in this molecule will be messed up, and should not be accessed later.

        Returns
        -------
        substructure : Molecule
        '''
        indexes = list(dict.fromkeys(indexes))

        if deepcopy:
            mol = copy.deepcopy(self)
        else:
            mol = self

        # assign atom id so that we can access id_atoms for connectivities
        for i, atom in enumerate(mol.atoms):
            atom.id = i

        # store residue information
        residues = list(dict.fromkeys(mol.atoms[i].residue for i in indexes))
        residue_name_atoms = []
        if len(residues) > 1:
            for residue in residues:
                atoms = [atom for atom in residue.atoms if atom.id in indexes]
                residue_name_atoms.append((residue.name, atoms))

        sub = Molecule()
        for i in indexes:
            sub.add_atom(mol.atoms[i], update_topology=False)

        ids_set = set(indexes)
        for conn in mol.bonds:
            if {a.id for a in conn.atoms} <= ids_set:
                sub._bonds.append(conn)
        for atom in sub.atoms:
            for i in reversed(range(len(atom.bonds))):
                if not {a.id for a in atom.bonds[i].atoms} <= ids_set:
                    atom._bonds.pop(i)
        for conn in mol.angles:
            if {a.id for a in conn.atoms} <= ids_set:
                sub._angles.append(conn)
        for conn in mol.dihedrals:
            if {a.id for a in conn.atoms} <= ids_set:
                sub._dihedrals.append(conn)
        for conn in mol.impropers:
            if {a.id for a in conn.atoms} <= ids_set:
                sub._impropers.append(conn)

        # reconstruct residues
        if len(residues) > 1:
            for resname, atoms in residue_name_atoms:
                sub.add_residue(resname, atoms, refresh_residues=False)
            sub.refresh_residues()
        else:
            sub.name = residues[0].name

        return sub

    @staticmethod
    def merge(molecules, deepcopy=True):
        '''
        Merge several molecules into a single molecule.


        Parameters
        ----------
        molecules : list of Molecule
        deepcopy: bool
            If True, the molecules will be deep-copied, and be intact after the mergence.
            Otherwise, the atoms of the merged molecule and of the original molecules will be the same objects.
            Then the original molecules will be unusable.

        Returns
        -------
        merged : Molecule
        '''
        merged = Molecule()
        for mol in molecules:
            if deepcopy:
                m_copy = copy.deepcopy(mol)
            else:
                m_copy = mol
            # should always call `add_atom()` instead of manipulating `_atoms` directly
            for atom in m_copy.atoms:
                merged.add_atom(atom)
            merged._bonds.extend(m_copy._bonds)
            merged._angles.extend(m_copy._angles)
            merged._dihedrals.extend(m_copy._dihedrals)
            merged._impropers.extend(m_copy._impropers)

            # all atoms goes to the default residue after `add_atom`, but the old residue still holds the atom list
            for residue in m_copy.residues:
                merged.add_residue(residue.name, residue.atoms, refresh_residues=False)

        merged.refresh_residues()

        return merged

    def split(self, consecutive=False):
        '''
        Split the molecule into smaller pieces based on bond network.

        The atoms in each piece will preserve the original order.
        However, the atoms at the end of original molecule may end up in a piece in the beginning,
        causing the order of all atoms in all the pieces different from original order.
        To avoid this, set consecutive to True.
        In this case, it will make sure all atoms in front pieces will have atom id smaller than atoms in back pieces.

        Residue information will be reconstructed for each piece

        Parameters
        ----------
        consecutive : bool

        Returns
        -------
        molecules : list of Molecule
        '''
        adjacency = self.get_adjacency_matrix()
        if consecutive:
            clusters = find_clusters_consecutive(self.atoms, lambda a1, a2: adjacency[a1.id_in_mol][a2.id_in_mol])
        else:
            clusters = find_clusters(self.atoms, lambda a1, a2: adjacency[a1.id_in_mol][a2.id_in_mol])

        for cluster in clusters:
            cluster.sort()
        clusters.sort()

        mol = copy.deepcopy(self)
        pieces = [mol.get_sub_molecule(ids, deepcopy=False) for ids in clusters]

        return pieces

    def split_residues(self):
        '''
        Split the molecule into smaller pieces. Each piece will be made of one residue.

        Make sure that there is no inter-residue bonds/angles/dihedrals/impropers.

        Returns
        -------
        molecules : list of Molecule
        '''
        pieces = []
        mol = copy.deepcopy(self)
        for residue in mol.residues:
            sub = Molecule(residue.name)
            for atom in residue.atoms:
                sub.add_atom(atom)
            pieces.append(sub)

        for conn in mol.bonds:
            if len({atom.molecule for atom in conn.atoms}) > 1:
                raise Exception(f'Inter-residue bond {conn} in {self}')
            conn.atom1.molecule._bonds.append(conn)
        for conn in mol.angles:
            if len({atom.molecule for atom in conn.atoms}) > 1:
                raise Exception(f'Inter-residue angle {conn} in {self}')
            conn.atom1.molecule._angles.append(conn)
        for conn in mol.dihedrals:
            if len({atom.molecule for atom in conn.atoms}) > 1:
                raise Exception(f'Inter-residue dihedral {conn} in {self}')
            conn.atom1.molecule._dihedrals.append(conn)
        for conn in mol.impropers:
            if len({atom.molecule for atom in conn.atoms}) > 1:
                raise Exception(f'Inter-residue improper {conn} in {self}')
            conn.atom1.molecule._impropers.append(conn)

        return pieces



import numpy as np
from .atom import Atom
from .geometry import periodic_distance, periodic_angle, periodic_dihedral

__all__ = [
    'Bond',
    'Angle',
    'Dihedral',
    'Improper',
]


class Bond():
    '''
    A bond between two atoms.

    Unlike BondTerm in ForceField, the atoms are not sorted in connectivity
    because the properties of atoms (name, type, etc...) are prone to changing.
    It's better to sort them when compare bonds or match force field parameters.

    Parameters
    ----------
    atom1 : Atom
    atom2 : Atom
    order : int, Optional

    Attributes
    ----------
    atom1 : Atom
    atom2 : Atom
    order : int
        The integer bond order. See :class:`Bond.Order`
    '''

    class Order:
        '''
        Enumerator of integer bond orders

        The bonds in aromatic rings are in kekulized form.
        '''

        #: unspecified bond order
        UNSPECIFIED = 0
        #: order of single bond
        SINGLE = 1
        #: order of double bond
        DOUBLE = 2
        #: order of triple bond
        TRIPLE = 3

    def __init__(self, atom1, atom2, order=Order.UNSPECIFIED):
        self.atom1 = atom1
        self.atom2 = atom2
        self._order = order

    def __repr__(self):
        return f'<Bond: {self.name} {self.order}>'

    def equals(self, other):
        '''
        Check if two bonds represent the same connectivity.
        Return True if they contain the identical atoms, regardless of the sequence and bond order.

        Parameters
        ----------
        other : Bond

        Returns
        -------
        equal : bool
        '''
        if type(other) != Bond:
            return False
        return {self.atom1, self.atom2} == {other.atom1, other.atom2}

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, val):
        if val != self._order and self.atom1.molecule is not None:
            self.atom1.molecule._is_rdmol_valid = False
        self._order = val

    @property
    def name(self):
        '''
        Name of this bond

        Returns
        -------
        name : str
        '''
        return '%s-%s' % (self.atom1.name, self.atom2.name)

    @property
    def atoms(self):
        '''
        The atoms forming this bond

        Returns
        -------
        atoms : tuple of Atom
        '''
        return self.atom1, self.atom2

    @property
    def is_drude(self):
        '''
        Whether or not this bond is a Drude dipole bond

        Returns
        -------
        is : bool
        '''
        return self.atom1.is_drude or self.atom2.is_drude

    def evaluate(self, cell=None):
        '''
        Evaluate the length of this bond

        Parameters
        ----------
        box : UnitCell, Optional

        Returns
        -------
        value : float
        '''
        if cell:
            return periodic_distance(self.atom1.position, self.atom2.position, cell.size)

        delta = self.atom2.position - self.atom1.position
        return float(np.sqrt(delta.dot(delta)))

    @property
    def rdbond(self):
        '''
        The `rdkit.Chem.Bond` object associated with this bond.

        Returns
        -------
        rdbond : rdkit.Chem.Bond
        '''
        return self.atom1.molecule.rdmol.GetBondBetweenAtoms(self.atom1.id_in_mol, self.atom2.id_in_mol)

    @property
    def is_aromatic(self):
        return self.rdbond.GetIsAromatic()

    @property
    def is_conjugated(self):
        return self.rdbond.GetIsConjugated()

    @property
    def is_rotatable(self):
        if self.order != Bond.Order.SINGLE:
            return False
        if len(self.atom1.bonds) == 1 or (len(self.atom2.bonds)) == 1:
            return False

        return not self.rdbond.IsInRing()


class Angle():
    '''
    A angle between three atoms. The second atom is the central atom.

    Parameters
    ----------
    atom1 : Atom
    atom2 : Atom
    atom3 : Atom

    Attributes
    ----------
    atom1 : Atom
    atom2 : Atom
    atom3 : Atom
    '''

    def __init__(self, atom1, atom2, atom3):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3

    def __repr__(self):
        return '<Angle: %s>' % self.name

    def equals(self, other):
        '''
        Check if two angles represent the same connectivity.
        Return True if they contain the identical side atoms and center atom, regardless of the sequence of side atoms.

        Parameters
        ----------
        other : Angle

        Returns
        -------
        equal : bool
        '''
        if type(other) != Angle:
            return False
        if self.atom2 != other.atom2:
            return False
        return {self.atom1, self.atom3} == {other.atom1, other.atom3}

    @property
    def name(self):
        '''
        Name of this angle

        Returns
        -------
        name : str
        '''
        return '%s-%s-%s' % (self.atom1.name, self.atom2.name, self.atom3.name)

    @property
    def atoms(self):
        '''
        The atoms forming this angle

        Returns
        -------
        atoms : tuple of Atom
        '''
        return self.atom1, self.atom2, self.atom3

    @property
    def bonds(self):
        '''
        The two bonds between each side atom and the center atom

        Returns
        -------
        bonds : tuple of Bond
        '''
        _bond12 = Bond(self.atom1, self.atom2)
        _bond23 = Bond(self.atom2, self.atom3)
        try:
            bond12 = next(b for b in self.atom2.bonds if b.equals(_bond12))
        except StopIteration:
            raise Exception(f'{_bond12} not found in {self.atom2.molecule}')
        try:
            bond23 = next(b for b in self.atom2.bonds if b.equals(_bond23))
        except StopIteration:
            raise Exception(f'{_bond23} not found in {self.atom2.molecule}')

        return bond12, bond23

    def evaluate(self, cell=None):
        '''
        Evaluate the value of this angle in unit of radian

        Parameters
        ----------
        cell : UnitCell, Optional

        Returns
        -------
        value : float
        '''
        if cell:
            return periodic_angle(self.atom1.position, self.atom2.position, self.atom3.position, cell)

        vec1 = self.atom1.position - self.atom2.position
        vec2 = self.atom3.position - self.atom2.position
        cos = vec1.dot(vec2) / np.sqrt(vec1.dot(vec1) * vec2.dot(vec2))
        return float(np.arccos(np.clip(cos, -1, 1)))


class Dihedral():
    '''
    A dihedral between four atoms.

    Parameters
    ----------
    atom1 : Atom
    atom2 : Atom
    atom3 : Atom
    atom4 : Atom

    Attributes
    ----------
    atom1 : Atom
    atom2 : Atom
    atom3 : Atom
    atom4 : Atom
    '''

    def __init__(self, atom1, atom2, atom3, atom4):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.atom4 = atom4

    def __repr__(self):
        return '<Dihedral: %s>' % self.name

    def equals(self, other):
        '''
        Check if two dihedrals represent the same connectivity.
        Dihedrals with reversed sequence are considered as equal. E.g., i-j-k-l and l-k-j-i are the same.

        Parameters
        ----------
        other : Dihedral

        Returns
        -------
        equal : bool
        '''
        if type(other) != Dihedral:
            return False
        return (self.atom1 == other.atom1 and self.atom2 == other.atom2 and
                self.atom3 == other.atom3 and self.atom4 == other.atom4) \
               or (self.atom1 == other.atom4 and self.atom2 == other.atom3 and
                   self.atom3 == other.atom2 and self.atom4 == other.atom1)

    @property
    def name(self) -> str:
        '''
        Name of this dihedral

        Returns
        -------
        name : str
        '''
        return '%s-%s-%s-%s' \
               % (self.atom1.name, self.atom2.name, self.atom3.name, self.atom4.name)

    @property
    def atoms(self):
        '''
        The atoms forming this dihedral

        Returns
        -------
        atoms : tuple of Atom
        '''
        return self.atom1, self.atom2, self.atom3, self.atom4

    @property
    def bonds(self):
        '''
        The three bonds in this dihedral

        Returns
        -------
        bonds : tuple of Bond
        '''
        _bond12 = Bond(self.atom1, self.atom2)
        _bond23 = Bond(self.atom2, self.atom3)
        _bond34 = Bond(self.atom3, self.atom4)
        mol = self.atom1.molecule
        try:
            bond12 = next(b for b in self.atom1.bonds if b.equals(_bond12))
        except StopIteration:
            raise Exception(f'{_bond12} not found in {mol}')
        try:
            bond23 = next(b for b in self.atom2.bonds if b.equals(_bond23))
        except StopIteration:
            raise Exception(f'{_bond23} not found in {mol}')
        try:
            bond34 = next(b for b in self.atom3.bonds if b.equals(_bond34))
        except StopIteration:
            raise Exception(f'{_bond34} not found in {mol}')

        return bond12, bond23, bond34

    @property
    def angles(self):
        '''
        The two angles in this dihedral

        Returns
        -------
        angles : tuple of Angle
        '''
        _angle123 = Angle(self.atom1, self.atom2, self.atom3)
        _angle234 = Angle(self.atom2, self.atom3, self.atom4)
        mol = self.atom1.molecule
        try:
            angle123 = next(a for a in mol.angles if a.equals(_angle123))
        except StopIteration:
            raise Exception(f'{_angle123} not found in {mol}')
        try:
            angle234 = next(a for a in mol.angles if a.equals(_angle234))
        except StopIteration:
            raise Exception(f'{_angle234} not found in {mol}')

        return angle123, angle234

    def evaluate(self, cell=None):
        '''
        Evaluate the value of this dihedral in unit of radian

        Parameters
        ----------
        cell : UnitCell, Optional

        Returns
        -------
        value : float
        '''
        if cell:
            return periodic_dihedral(self.atom1.position, self.atom2.position, self.atom3.position, self.atom4.position,
                                     cell.size)

        vec1 = self.atom2.position - self.atom1.position
        vec2 = self.atom3.position - self.atom2.position
        vec3 = self.atom4.position - self.atom3.position
        n1 = np.cross(vec1, vec2)
        n2 = np.cross(vec2, vec3)
        cos = n1.dot(n2) / np.sqrt(n1.dot(n1) * n2.dot(n2))
        value = float(np.arccos(np.clip(cos, -1, 1)))
        sign = 1 if vec1.dot(n2) >= 0 else -1
        return sign * value


class Improper():
    '''
    An improper between four atoms. The first atom is the central atom.

    Parameters
    ----------
    atom1 : Atom
    atom2 : Atom
    atom3 : Atom
    atom4 : Atom

    Attributes
    ----------
    atom1 : Atom
    atom2 : Atom
    atom3 : Atom
    atom4 : Atom
    '''

    def __init__(self, atom1, atom2, atom3, atom4):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.atom4 = atom4

    def __repr__(self):
        return '<Improper: %s>' % self.name

    def equals(self, other):
        '''
        Check if two impropers represent the same connectivity.
        Impropers are considered as equal if they have the same side atoms and central atom, regardless of the sequence of side atoms.
        So i-j-k-l and i-k-l-j are the same.

        Parameters
        ----------
        other : Improper

        Returns
        -------
        equal : bool
        '''
        if type(other) != Improper:
            return False
        if self.atom1 != other.atom1:
            return False
        return {self.atom2, self.atom3, self.atom4} == {other.atom2, other.atom3, other.atom4}

    @property
    def name(self):
        '''
        Name of this improper

        Returns
        -------
        name : str
        '''
        return '%s-%s-%s-%s' % (self.atom1.name, self.atom2.name, self.atom3.name, self.atom4.name)

    @property
    def atoms(self):
        '''
        The atoms forming this improper torsion

        Returns
        -------
        atoms : tuple of Atom
        '''
        return self.atom1, self.atom2, self.atom3, self.atom4

    def evaluate(self, cell=None):
        '''
        Evaluate the value of this improper torsion in unit of radian.
        The improper is defined as the angle between plane a1-a2-a3 and a2-a3-a4.

        Parameters
        ----------
        cell : UnitCell, Optional

        Returns
        -------
        value : float
        '''
        if cell:
            return periodic_dihedral(self.atom1.position, self.atom2.position, self.atom3.position, self.atom4.position,
                                     cell.size)

        vec1 = self.atom2.position - self.atom1.position
        vec2 = self.atom3.position - self.atom2.position
        vec3 = self.atom4.position - self.atom3.position
        n1 = np.cross(vec1, vec2)
        n2 = np.cross(vec2, vec3)
        cos = n1.dot(n2) / np.sqrt(n1.dot(n1) * n2.dot(n2))
        value = float(np.arccos(np.clip(cos, -1, 1)))
        sign = 1 if vec1.dot(n2) >= 0 else -1
        return sign * value
